"""

Mirantis MKE API interaction.

Two Metta plugins:

1. The Metta cli plugin allows CLI interaction with MKE client plugin
   from the command line tool.
2. The Metta client plugin allows interaction with the MKE API.

"""

import logging
import json
from typing import Dict, List
from enum import Enum

import requests
import toml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta_cli.base import CliBase, cli_output

logger = logging.getLogger('metta.contrib.metta_mirantis.client.mkeapi')

METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID = 'metta_mirantis_client_mke'
""" Mirantis MKE API Client plugin id """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class MKEAPICliPlugin(CliBase):
    """Metta CLI plugin which injects the MKE API Client CLI commands."""

    def fire(self):
        """Return a dict of commands for MKE API Clients."""
        if len(self.environment.fixtures.filter(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                exception_if_missing=False)) > 0:
            return {
                'contrib': {
                    'mke': MKEAPICliGroup(self.environment)
                }
            }

        return {}


class MKEAPICliGroup():
    """Metta CLI plugin which provides the MKE API Client CLI commands."""

    def __init__(self, environment: Environment):
        """Create new cli group object."""
        self.environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """Pick a matching fixture in case there are more than one."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                instance_id=instance_id)

        # Get the highest priority fixture
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)

    def info(self, instance_id: str = '', deep: bool = False):
        """Get info about a plugin."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        info.update(plugin.info(deep))

        return cli_output(info)

    def version(self, instance_id: str = ''):
        """Get MKE Version."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_version())

    def ping(self, instance_id: str = '', node: int = None):
        """Check if we can ping MKE."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return "OK" if plugin.api_ping(node) else "FAIL"

    def pingall(self, instance_id: str = ''):
        """Check if we can ping all of the nodes directly."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        ping = {}
        for index in range(0, plugin.host_count()):
            try:
                plugin.api_ping(index)
                # pylint: disable=protected-access
                ping[plugin._node_address(index)] = True
            except requests.exceptions.RequestException:
                # pylint: disable=protected-access
                ping[plugin._node_address(index)] = False

        return cli_output(ping)

    # pylint: disable=invalid-name
    def id(self, instance_id: str = ''):
        """Get auth id from MKE."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_id())

    def nodes(self, instance_id: str = '', node_id: str = ''):
        """List swarm nodes."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_nodes(node_id))

    def services(self, instance_id: str = '', service_id: str = ''):
        """List swarm services."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_services(service_id))

    def tasks(self, instance_id: str = '', task_id: str = ''):
        """List swarm tasks."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        if task_id:
            return cli_output(plugin.api_task(task_id))
        return cli_output(plugin.api_tasks())

    def tomlconfig(self, instance_id: str = ''):
        """Get the toml config."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_ucp_configtoml_get())

    def tomlconfig_set(self, table: str, key: str,
                       value: str, instance_id: str = ''):
        """Set a single toml config value."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin

        data = plugin.api_ucp_configtoml_get()
        data[table][key] = value
        return cli_output(plugin.api_ucp_configtoml_put(data))

    def metricsdiscovery(self, instance_id: str = ''):
        """Get the api metrics-discover."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_metrics_discovery())

    def metrics(self, instance_id: str = ''):
        """Get the toml config."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin.api_metrics_discovery())

    # pylint: disable=protected-access
    def auth(self, instance_id: str = ''):
        """Retrieve auth headersg."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return cli_output(plugin._auth_headers())


class MKENodeState(Enum):
    """MKE Node state in the node status API response."""

    UNKNOWN = 'unknown'
    DOWN = 'down'
    READY = 'ready'
    DISCONNECTED = 'disconnected'

    # pylint: disable=comparison-with-callable
    def match(self, compare: str) -> bool:
        """Allow for string comparisons."""
        return self.value == compare


# pylint: disable=too-many-instance-attributes
class MKEAPIClientPlugin:
    """Metta Client plugin for API Connections to MKE."""

    # pylint: disable=too-many-arguments"
    def __init__(self, environment: Environment, instance_id: str,
                 accesspoint: str, username: str, password: str, hosts: List[Dict]):
        """Create an MKE Client plugin instance.

        Parameters:
        -----------
        Standard Metta plugin parameters:

        environment (Environment) : env in which this plugin exists
            can be used to get access to config, other fixtures etc.
        instance_id (str) : string ID for this plugin to self-identify

        Plugin specific parameters:

        accesspoint (str) : API URL endpoint

        username / password (str/str) : API authentication credentials

        hosts (List[Dict]) : List of host definition dicts used to define hosts
            that can respond to API requests in the cluster.  This can be used
            to allow directly accessing the API through a specific host.
            The first node in the list is used as an API endpoint if no
            endpoint is specified.

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.accesspoint = accesspoint
        self.username = username
        self.password = password

        self.hosts = hosts
        """ List of hosts """

        if self.accesspoint is None:
            # use the first host as an accesspoint if none was delivered
            self.accesspoint = self._node_address(0)

        self.verify = False
        """ should we verify ssl certs """
        if not self.verify:
            # pylint: disable=no-member
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning)

        self.auth_token = None
        """ hold the bearer auth token created by ._auth_headers() (cache) """

    def host_count(self):
        """Return integer host count for MKE cluster."""
        return len(self.hosts)

    def info(self, deep: bool = False):
        """Return information about the plugin."""
        info = {
            'api': {
                'accesspoint': self.accesspoint,
                'username': self.username,
            },
            'hosts': self.hosts
        }

        if deep:
            info['id'] = self.api_id()
            info['version'] = self.api_version()
            info['api_info'] = self.api_info()

        return info

    def auth_header(self) -> Dict:
        """Retrieve the auth headers so you can do your own thing."""
        return json.loads(self._auth_headers())

    def api_ping(self, node: int = None) -> bool:
        """Check the API ping response."""
        if node is not None:
            endpoint = self._accesspoint_url('_ping', node=node)
        else:
            endpoint = self._accesspoint_url('_ping')

        with requests.get(endpoint, verify=self.verify) as response:
            response.raise_for_status()
            return response.ok

    def api_id(self) -> Dict:
        """Retrieve the API id."""
        with requests.get(self._accesspoint_url('id'), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_version(self) -> Dict:
        """Retrieve version."""
        with requests.get(self._accesspoint_url('version'), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_info(self) -> Dict:
        """Retrieve the API info."""
        with requests.get(self._accesspoint_url('info'), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_nodes(self, node_id: str = '') -> Dict:
        """Retrieve the API nodes."""
        endpoint = f'nodes/{node_id}' if node_id else 'nodes'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_services(self, service_id: str = '') -> Dict:
        """Retrieve the API services."""
        endpoint = f'services/{service_id}' if service_id else 'services'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_metrics_discovery(self):
        """Retrieve the API metrics."""
        endpoint = 'metricsdiscovery'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_tasks(self):
        """Retrieve the API tasks."""
        endpoint = 'tasks'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_task(self, task_id: str):
        """Retrieve the API tasks."""
        endpoint = f'tasks/{task_id}'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_ucp_configtoml_get(self):
        """Retrieve config toml as a struct."""
        endpoint = 'api/ucp/config-toml'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return toml.loads(response.text)

    def api_ucp_configtoml_put(self, data):
        """Send struct config as a toml string."""
        endpoint = 'api/ucp/config-toml'
        data_toml = toml.dumps(data)
        with requests.put(self._accesspoint_url(endpoint), headers=self._auth_headers(),
                          verify=self.verify, data=data_toml) as response:
            response.raise_for_status()
            return {'response': json.loads(response.content), 'data': data}

    def _auth_headers(self):
        """Get an auth token."""
        if self.auth_token is None:
            data = {
                "password": self.password,
                "username": self.username
            }
            with requests.post(self._accesspoint_url('auth/login'), data=json.dumps(data),
                               verify=self.verify) as response:
                response.raise_for_status()
                content = json.loads(response.content)
                self.auth_token = content['auth_token']

        return {'Authorization': f'Bearer {self.auth_token}'}

    def _accesspoint_url(self, endpoint: str = '', node: int = None):
        """Convert an endpoint into a full URL for an API Call.

        Pass in a sub-url endpoint and this will convert it into a full URL.
        You can request a specific host index if desired.

        Parameters:
        -----------
        endpoint (str) : API endpoint you are trying to access

        node (int) : If not None, then the API call should be directed to a
            specific node index from the list of hosts, as opposed to the
            generic accesspoint. This allows things like ping confirmation on
            specific host.

        """
        if node is None:
            target = self.accesspoint
        else:
            target = self._node_address(node)

        return f"https://{target.rstrip('/')}/{endpoint.lstrip('/')}"

    def _node_address(self, node: int = 0):
        """Get the ip address from the node for the node index."""
        node_dict = self.hosts[node]
        if 'address' in node_dict:
            return node_dict['address']
        if 'ssh' in node_dict:
            return node_dict['ssh']['address']
        if 'winrm' in node_dict:
            return node_dict['winrm']['address']

        raise ValueError(f"No node address could be found for the node {node}")



"""

Global mutables.

Keep all package global variables in this module.

"""
from typing import Dict

from mirantis.testing.metta.environment import Environment

all_environments: Dict[str, Environment] = {}
""" Keep a Dict of all created environments for introspection """



"""

Ansible provisioner

"""

from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .provisioner import (AnsibleProvisionerPlugin, METTA_ANSIBLE_PROVISIONER_PLUGIN_ID,
                          ANSIBLE_PROVISIONER_CONFIG_LABEL)
from .cli import AnsibleCliPlugin, METTA_ANSIBLE_CLI_PLUGIN_ID


@Factory(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_ANSIBLE_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_ansible(
        environment: Environment,
        instance_id: str = "",
        label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT):
    """ create an metta provisionersss dict plugin """
    return AnsibleProvisionerPlugin(environment, instance_id, label, base)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_ANSIBLE_CLI_PLUGIN_ID)
def metta_ansible_factory_cli_ansible(
        environment: Environment, instance_id: str = ''):
    """ create an info cli plugin """
    return AnsibleCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """ METTA_Terraform bootstrap

    Currently we only use this to import plugins.

    Parameters:
    -----------

    env (Environment) : an environment which should have validation config added to.

    """



"""

Mirantis MSR API interaction.

Two Metta plugins:

1. The Metta cli plugin allows CLI interaction with MSR client plugin
   from the command line tool.
2. The Metta client plugin allows interaction with the MSR API.

"""

import logging
import json
from typing import List, Dict
from enum import Enum

import requests

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta_cli.base import CliBase

logger = logging.getLogger('metta.contrib.metta_mirantis.client.msrapi')


METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID = 'metta_mirantis_client_msr'
""" Mirantis MSR APIP Client plugin id """

# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods


class MSRAPICliPlugin(CliBase):
    """Metta CLI plugin for injecting MSR API Client commands into the cli."""

    def fire(self):
        """Return any MSR CLI command groups."""
        if self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'msr': MSRAPICliGroup(self.environment)
                }
            }

        return {}


class MSRAPICliGroup():
    """MSR API Client CLI commands."""

    def __init__(self, environment: Environment):
        """Create a new MSR CLI command group."""
        self.environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """Pick a matching fixture in case there are more than one."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                instance_id=instance_id)

        # Get the highest priority fixture
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)

    def info(self, instance_id: str = '', deep: bool = False):
        """Get info about a plugin."""
        fixture = self._select_fixture(instance_id=instance_id)

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                info.update(fixture.plugin.info(deep))

        return json.dumps(info, indent=2)

    def ping(self, instance_id: str = '', node: int = None):
        """Ping an MSR replica."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_ping(node=node), indent=2)

    def pingall(self, instance_id: str = ''):
        """Check if we can ping all of the nodes directly."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        ping = {}
        for index in range(0, plugin.host_count()):
            try:
                plugin.api_ping(index)
                # pylint: disable=protected-access
                ping[plugin._node_address(index)] = True
            except requests.exceptions.RequestException:
                # pylint: disable=protected-access
                ping[plugin._node_address(index)] = False

        return json.dumps(ping, indent=2)

    def health(self, instance_id: str = '', node: int = None):
        """Get the MSR health api response."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_health(node=node), indent=2)

    def nginx_status(self, instance_id: str = '', node: int = None):
        """Get the MSR nginsx status api response."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_nginx_status(node=node), indent=2)

    def version(self, instance_id: str = ''):
        """Get MSR cluster version."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_version(), indent=2)

    def status(self, instance_id: str = ''):
        """Get cluster status."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_status(), indent=2)

    def features(self, instance_id: str = ''):
        """Get features list."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_features(), indent=2)

    def alerts(self, instance_id: str = ''):
        """Get alerts list."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_alerts(), indent=2)


class MSRReplicaHealth(Enum):
    """MSR replica health values (in the cluster status API response)."""

    OK = 'OK'

    def match(self, compare: str) -> bool:
        """Allow for string comparisons."""
        # pylint: disable=comparison-with-callable
        return self.value == compare


# pylint: disable=too-many-instance-attributes
class MSRAPIClientPlugin:
    """Client for API Connections to MSR."""

    # pylint: disable=too-many-arguments"
    def __init__(self, environment: Environment, instance_id: str, accesspoint: str,
                 username: str, password: str, hosts: List[Dict], api_version: str = 'v0'):
        """Create new MSR Client API plugin.

        Parameters:
        -----------
        Standard Metta plugin parameters:

        environment (Environment) : env in which this plugin exists
            can be used to get access to config, other fixtures etc.
        instance_id (str) : string ID for this plugin to self-identify

        Plugin specific parameters:

        accesspoint (str) : API URL endpoint

        username / password (str/str) : API authentication credentials

        hosts (List[Dict]) : List of host definition dicts used to define hosts
            that can respond to API requests in the cluster.  This can be used
            to allow directly accessing the API through a specific host.
            The first node in the list is used as an API endpoint if no
            endpoint is specified.

        """
        self.environment = environment
        """ Environemnt in which this plugin exists."""
        self.instance_id = instance_id
        """ Unique id for this plugin instance."""

        self.accesspoint = accesspoint
        self.username = username
        self.password = password

        self.hosts = hosts
        """ List of hosts """

        if not self.accesspoint:
            # use the first host as an accesspoint if none was delivered
            self.accesspoint = self._node_address(0)

        self.version = api_version
        """ used to build the api url """

        self.verify = False
        """ should we verify ssl certs : default to NO """
        if not self.verify:
            # pylint: disable=no-member
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning)

        self.auth_token = None
        """ hold the bearer auth token if created by ._auth_headers() """

    def host_count(self):
        """Return integer host count for MSR cluster."""
        return len(self.hosts)

    def info(self, deep: bool = False):
        """Return information about the plugin."""
        info = {
            'api': {
                'accesspoint': self.accesspoint,
                'username': self.username,
            },
            'hosts': self.hosts
        }

        if deep:
            info['version'] = self.api_version()
            info['features'] = self.api_features()
            info['status'] = self.api_status()

        return info

    def api_ping(self, node: int = None) -> bool:
        """Check the API ping response."""
        endpoint = self._accesspoint_url('_ping', root_api=True, node=node)

        with requests.get(endpoint, verify=self.verify) as response:
            response.raise_for_status()
            return response.ok

    def api_health(self, node: int = None):
        """Check the API ping response."""
        endpoint = self._accesspoint_url('health', root_api=True, node=node)

        with requests.get(endpoint, auth=self._api_auth(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_nginx_status(self, node: int = None):
        """Check the API ping response.

        Returns:
        --------
        parsed string contents of the response

        """
        endpoint = self._accesspoint_url(
            'nginx_status', root_api=True, node=node)

        with requests.get(endpoint, auth=self._api_auth(),
                          headers={'content-type': 'application/json'},
                          verify=self.verify) as response:
            response.raise_for_status()
            # @TODO should we parse this?
            return response.content.decode("utf-8")

    def api_version(self) -> Dict:
        """Retrieve version."""
        with requests.get(self._accesspoint_url('admin/version'), auth=self._api_auth(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_status(self) -> Dict:
        """Retrieve status from the api."""
        with requests.get(self._accesspoint_url('meta/cluster_status'), auth=self._api_auth(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_features(self) -> Dict:
        """Retrieve features list from the api."""
        with requests.get(self._accesspoint_url('meta/features'), auth=self._api_auth(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_alerts(self) -> Dict:
        """Retrieve alerts list from the api."""
        with requests.get(self._accesspoint_url('meta/alerts'), auth=self._api_auth(),
                          verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def _api_auth(self):
        """Get the requests auth handler."""
        return requests.auth.HTTPBasicAuth(self.username, self.password)

    def _accesspoint_url(self, endpoint: str = '',
                         root_api: bool = False, node: int = None) -> str:
        """Convert an endpoint into a full URL for an API Call.

        Pass in a sub-url endpoint and this will convert it into a full URL.
        Most endpoints will have the api/v0/ prefix added (unless it is a root
        API call) and will get passed on to the configured accesspoint.

        You can request a specific host index if desired.

        Parameters:
        -----------
        endpoint (str) : API endpoint you are trying to access

        root_api (bool) : If True, then the URL should be for a root API call
            which doesn't have any version info

        node (int) : If not None, then the API call should be directed to a specific
            node index from the list of hosts, as opposed to the generic accesspoint.
            This allows things like ping confirmation on specific host.

        Returns:
        --------
        String URL for API endpoint

        """
        if node is None:
            target = self.accesspoint
        else:
            target = self._node_address(node)

        if root_api:
            return f"https://{target}/{endpoint}"

        return f"https://{target}/api/{self.version}/{endpoint}"

    def _node_address(self, node: int = 0) -> str:
        """Get the ip address from the node for the node index.

        We get this information from the list of nodes.  This list of nodes can
        come in various forms, because the primary consumer here is a list of
        nodes from launchpad yaml, and there is variation in that syntax.
        It's not great coding, but we allow for variability here to handle
        differing incoming formats.

        Returns:
        --------
        String node address

        """
        node_dict = self.hosts[node]
        if 'address' in node_dict:
            return node_dict['address']
        if 'ssh' in node_dict:
            return node_dict['ssh']['address']
        if 'winrm' in node_dict:
            return node_dict['winrm']['address']

        raise ValueError(f"No node address could be found for the node {node}")



"""

Mirantis preset functionality.

The preset functionality allows reuse of common sets of configuration that is
commonly used for various testcases.  This allows centralized definition of
things like release versions of Mirantis products.

To consume the presets, you define some config in the mirantis label, which
will tell this code to include more config, in the package root config folder.

"""
import os.path
import logging
import pkg_resources

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH
from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta_mirantis.presets")

METTA_PRESET_CONFIG_LABEL = "mirantis"
""" This config label loaded to interpret config in add_preset_config """
METTA_CONFIG_SOURCE_INSTANCE_ID_PREFIX = "metta_mirantis-preset"
""" prefix for all config that we add here """
METTA_CONFIG_CONFIG_PRESET_BASE = 'presets'
""" preset configuration will be found under this config .get() key """
METTA_PRESET_DEFAULT_PRIORITY = 60
""" Default priority to use for preset config sources """
METTA_PRESET_PACKAGE = 'mirantis.testing.metta_mirantis'


def preset_config():
    """List of available presets that you can include in presets.

    In load("metta").get('presets')

    (key, preset_root_path):
        key: config key that will find the preset value under the base value
            from METTA_CONFIG_CONFIG_PRESET_BASE
        path: preset root path that should contains a path matching the value
            which will be added as a source config path
        priority delta: preset relative priority compared to other presets
    """
    return [
        ("variation", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/variation"), 0),
        ("cluster", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/cluster"), 1),
        ("platform", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/platform"), 1),
        ("release", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/release"), 1)
    ]


def add_preset_config(environment: Environment,
                      priority=METTA_PRESET_DEFAULT_PRIORITY):
    """Read metta config and interpret it for modifying the config.

    Interprests config.load("metta") to determine if any `presets` config
    sources should be added.

    The following presets are currently available:

    1. variation : see ./config/variations : kind of access to a larger config
        pattern that we use at Mirantis.  An example is `lat` which refers to
        using launchpad & terraform with our common config.

    2. cluster : @see ./config/cluster : config for different cluster sizes
    3. release : @see ./config/release : config for different mirantis product
        releases
    4. platform : @see ./config/platform : config for different os platforms

    In all cases, the pattern is that a key from config.load("metta") is
    used to pull a preset name, and the matching preset subpath is added to the
    config object as a path source.

    The presets for #2, #3 and # are focused on overriding `variables` (label)
    which should be consumed by whichever variation you use.

    Parameter:
    ----------
    config (configerus.config.Config) : a configerus Config object which will
        have config sources added

    priority (int) : optionally override the source plugin priority.

    Throws:
    -------
    Can throw a KeyError if a preset id requested that does not exist

    """
    # Now we start to process `presets`
    # To do this we load the `metta` config label and look for preset
    # keys in the loaded config.
    # For each preset key found we try to add a path config source for the
    # value
    try:
        metta_config = environment.config.load(METTA_PRESET_CONFIG_LABEL)
    except KeyError:
        logger.debug("metta Presets found no usable config")
        return

    for preset in preset_config():
        # For each of the available presets, look in the loaded metta
        # config for values matching the preset key.  if the key is found then
        # we try to match a path in the metta/config/{key}/{value}
        # and add the path as a config source.

        (preset_key, preset_root_path, preset_priority_delta) = preset
        preset_priority = priority + preset_priority_delta
        preset_value = metta_config.get(
            [METTA_CONFIG_CONFIG_PRESET_BASE, preset_key], default='')

        if preset_value:
            prefix = METTA_CONFIG_SOURCE_INSTANCE_ID_PREFIX
            preset_key = preset_value.replace('/', '_')
            preset_instance_id = f"{prefix}-{preset_key}"
            # quick check to see if we've already added this preset.
            if not environment.config.has_source(preset_instance_id):
                # build a preset config path and add it as a source if it
                # exists
                preset_full_path = os.path.join(preset_root_path, preset_value)
                if os.path.isdir(preset_full_path):
                    logger.debug("Using metta preset %s:%s => %s", preset_key, preset_value,
                                 preset_full_path)
                    environment.config.add_source(
                        PLUGIN_ID_SOURCE_PATH,
                        preset_instance_id,
                        preset_priority).set_path(preset_full_path)
                else:
                    raise KeyError(f"metta doesn't have a preset '{preset_key}:{preset_value}'")

        else:
            logger.debug("No metta preset selected for %s", preset_key)



"""

Metta CLI : User plugin

Examine and declare user configuration for Metta

"""
import logging
import os
import json

import appdirs
import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_cli.base import CliBase

from .common_config import METTA_COMMON_APP_NAME

logger = logging.getLogger('metta.cli.user')


METTA_PLUGIN_ID_CLI_USER = 'user'
""" cli plugin_id for the user plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class UserCliPlugin(CliBase):
    """Metta cli command generator for commands related to the current user."""

    def fire(self):
        """Return a dict of commands."""
        return {
            'user': UserGroup(self.environment)
        }


class UserGroup():
    """Commands relating to the current user.

    These commands can be used to manage the user/system settings for defining functionality that
    should apply to all projects, at the user level.

    """

    def __init__(self, environment: Environment):
        """Add environment to command group."""
        self._environment = environment

    def info(self):
        """Output any user related information."""
        user_config = self._environment.config.load('user')
        user_config_dir = _user_path()

        info = {
            'config': user_config.get(default={}),
            'system': {
                'path': user_config_dir,
                'exists': os.path.isdir(user_config_dir)
            }
        }

        return json.dumps(info, indent=2)

    # pylint: disable=no-self-use
    def init(self):
        """Make a User config source that will be global for this system."""
        user_config_dir = _user_path()
        if _user_path_exists():
            raise RuntimeError(
                f"User has already been initialized on this system: {user_config_dir}")

        os.mkdir(user_config_dir)

    # pylint: disable=no-self-use
    def set_id(self, uid: str):
        """Set the user id for the user on the system."""
        if not _user_path_exists():
            raise RuntimeError(
                "No user configuration has been initialized on this system.  Run `init` first")

        user_config_dir = _user_path()

        config_path_user = os.path.join(user_config_dir, 'user.yml')
        try:
            with open(config_path_user, 'r') as user_config_file:
                user_config = yaml.load(user_config_file)
        except FileNotFoundError:
            user_config = {}

        user_config['id'] = uid

        with open(config_path_user, 'w') as user_config_file:
            yaml.dump(user_config, user_config_file)


def _user_path():
    """Get path to user specific configuration."""
    # a user config path (like ~/.config/metta) may contain config
    return appdirs.user_config_dir(METTA_COMMON_APP_NAME)


def _user_path_exists():
    """Perform boolean check on whether or not the user config path exists on the system."""
    # a user config path (like ~/.config/metta) may contain config
    user_config_dir = _user_path()
    return os.path.isdir(user_config_dir)



"""

Ansible METTA provisioner plugin

This package is a WIP and currently doesn't really do anything.

"""

import logging
from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.provisioner import ProvisionerBase

logger = logging.getLogger('metta.contrib.provisioner:ansible')

METTA_ANSIBLE_PROVISIONER_PLUGIN_ID = 'metta_ansible'
""" Ansible provisioner plugin id """

ANSIBLE_PROVISIONER_CONFIG_LABEL = 'ansible'
""" config label loading the ansible config """
ANSIBLE_PROVISIONER_CONFIG_PLAN_PATH_KEY = 'plan.path'
""" config key for the ansible plan path """
ANSIBLE_PROVISIONER_CONFIG_STATE_PATH_KEY = 'state.path'
""" config key for the ansible state path """
ANSIBLE_PROVISIONER_CONFIG_VARS_KEY = 'vars'
""" config key for the ansible vars Dict, which will be written to a file """
ANSIBLE_PROVISIONER_CONFIG_VARS_PATH_KEY = 'vars_path'
""" config key for the ansible vars file path, where the plugin will write to """
ANSIBLE_PROVISIONER_DEFAULT_VARS_FILE = 'metta_ansible.tfvars.json'
""" Default vars file if none was specified """
ANSIBLE_PROVISIONER_DEFAULT_STATE_SUBPATH = 'metta-state'
""" Default vars file if none was specified """


class AnsibleProvisionerPlugin(ProvisionerBase):
    """ Ansible provisioner plugin

    Provisioner plugin that allows control of and interaction with a ansible
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a ansible binary, so you have to install
       ansible in the environment

    ## Usage

    @NOTE his plugin ist still very much a stub, and implements no real functionality.

    """

    def __init__(self, environment, instance_id,
                 label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties

        Interpret provided config and configure the object with all of the needed
        pieces for executing ansible commands

        """
        super().__init__(environment, instance_id)

        logger.info("Preparing Ansible setting")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def info(self):
        """ get info about a provisioner plugin """
        return {
            'plugin': {
                'config_label': self.config_label,
                'config_base': self.config_base
            }
        }

    def prepare(self):
        """ Prepare the provisioner to apply resources

        Initial Provisioner plugin is expected to be of very low cost until
        prepare() is executed.  At this point the plugin should load any config
        and perform any validations needed.
        The plugin should not create any resources but it is understood that
        there may be a cost of preparation.

        Provisioners are expected to load a lot of config to self-program.
        Because of this, they allow passing of a configerus label for .load()
        and a base for .get() in case there is an alterante config source
        desired.

        """

    def apply(self):
        """ bring a cluster to the configured state

        Run an ansible playbook to execute functionality on a running cluster.

        @NOTE his plugin ist still very much a stub, and implements no real functionality.

        """

    def destroy(self):
        """ remove all resources created for the cluster

        Run an aansible palybook to remove any implemented changes

        @NOTE his plugin ist still very much a stub, and implements no real functionality.

        """



"""

Metta cli plugin for testkit.

Mainly provides functionality for interacting with the testkit provisioner.

"""
import logging


from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import CliBase, cli_output

logger = logging.getLogger('metta.cli.testkit')

METTA_PLUGIN_ID_TESTKIT_CLI = 'metta_testkit'
""" metta plugin id for the testkit cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class TestkitCliPlugin(CliBase):
    """Fire command/group generator for testkit commands."""

    def fire(self):
        """Return a dict of commands."""
        if self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_PLUGIN_ID_TESTKIT_CLI,
                exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'testkit': TestkitGroup(self.environment)
                }
            }

        return {}


class TestkitGroup():
    """Base Fire command group for testkit commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """Pick a matching provisioner."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_PLUGIN_ID_TESTKIT_CLI,
                instance_id=instance_id)

        # Get the highest priority provisioner
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_PLUGIN_ID_TESTKIT_CLI)

    def info(self, provisioner: str = '', deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                info.update(fixture.plugin.info(True))

        return cli_output(info)

    def prepare(self, provisioner: str = ''):
        """Run the provisioner prepare : which for testkit currently does nothing."""

    def apply(self, provisioner: str = ''):
        """Run the provisioner apply, which runs testkit create."""
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        return cli_output(plugin.apply())

    def destroy(self, provisioner: str = ''):
        """Run the provisioner destroy, which runs testkit system rm."""
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        return cli_output(plugin.destroy())



"""

Helm workload Metta plugin

As a workload, using a kube_api client, manage helm charts

"""
from typing import Any, List, Dict
import logging
import subprocess
from enum import Enum

import yaml

from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import WorkloadBase

from .kubeapi_client import METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger('metta.contrib.kubernetes.workload.helm')

METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD = 'metta_kubernetes_helm'
""" workload plugin_id for the metta_kubernetes helm plugin """


KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL = 'kubernetes'
""" default config label used to load the workload """
KUBERNETES_HELM_WORKLOAD_CONFIG_BASE = 'workload.helm'
""" default config key for creating a workload object """

KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VERSION = 'version'
""" config key for helm version """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS = 'repos'
""" config key for helm repos dict which need to be added """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART = 'chart'
""" config key for helm chart, either local path or http(s) """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET = 'set'
""" config key for helm chart values """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES = 'values'
""" config key for helm chart values that should be put into a values file """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH = 'file'
""" config key for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH = 'values.yml'
""" config default value for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE = 'namespace'
""" config key for namespace to install to """

KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE = 'default'
""" default namespace to install to if no namespace was passed """

KUBERNETES_HELM_WORKLOAD_DEFAULT_BIN = 'helm'
""" default helm executble path """
KUBERNETES_HELM_WORKLOAD_DEFAULT_WORKINGDIR = '.'
""" default helm working dir for subprocess """


class KubernetesHelmWorkloadPlugin(WorkloadBase):
    """Kubernetes workload class."""

    def __init__(self, environment, instance_id, label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL,
                 base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receiv config pointers so that we can determine what
        helm workload to apply, and what values will be needed.

        Parameters:
        -----------
        config_file (str): String path to the kubernetes config file to use

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        try:
            client = fixtures.get_plugin(plugin_type=METTA_PLUGIN_TYPE_CLIENT,
                                         plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)
        except KeyError as err:
            raise NotImplementedError("Workload could not find the needed client: "
                                      f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}") from err

        workload_config = self.environment.config.load(self.config_label)

        kubeconfig = client.config_file

        namespace = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE],
            default=KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE)

        chart = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART])

        values_set = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET], default={})
        values = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES], default={})
        file = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH],
            default=KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH)

        repos = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS], default={})

        return KubernetesHelmV3WorkloadInstance(
            kubeconfig=kubeconfig, namespace=namespace, name=self.instance_id, repos=repos,
            chart=chart, values_set=values_set, values=values, file=file)

    def info(self):
        """Return dict data about this plugin for introspection."""
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'repos': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS]),
                    'chart': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART]),
                    'set': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET]),
                    'values': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES])
                },
                'required_fixtures': {
                    'kubernetes': {
                        'plugin_type': METTA_PLUGIN_TYPE_CLIENT,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


# This is how many vars it takes to program helm
# pylint: disable=too-many-instance-attributes
class KubernetesHelmV3WorkloadInstance:
    """A workload instance, which is a single active helm chart.

    A workload plugin can create many isntances.

    """

    # pylint: disable=too-many-arguments
    def __init__(self, kubeconfig, namespace, name, repos, chart, values_set: Dict[str, str] = None,
                 values: Dict[str, Any] = None,
                 helm_bin: str = KUBERNETES_HELM_WORKLOAD_DEFAULT_BIN,
                 file: str = KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH,
                 work_dir: str = KUBERNETES_HELM_WORKLOAD_DEFAULT_WORKINGDIR):
        """Set initial instance state.

        Parameters:
        -----------
        kubeconfig (str) : Path to the kubeinfo file

        repos (dict[str, str]) : name->path for repos to add

        chart (str) : local or http(s) path to a chart

        value_set (dict[str, str]) : simple helm chart values that can be set using
           the --set options on the command line

        values (dict[str, Any]) : complex helm chart value data that needs to
            be put into a file
        file (str) : path to where we should put values yaml file

        bin (str) : helm executable path
        dir (dir) : working dir to be used with subprocess

        """
        self.kubeconfig = kubeconfig
        self.name = name
        self.namespace = namespace
        self.repos = repos
        self.chart = chart

        self.set = values_set if values_set is not None else {}
        self.values = values if values is not None else {}
        self.file = file

        self.bin = helm_bin
        self.working_dir = work_dir

    def apply(self, wait: bool = True, debug: bool = False):
        """Apply the helm chart.

        To make the helm apply method reusable, we always run an upgrade with
        the --install flag, which w orks for both the first install, and any
        upgrades run after install

        Parameters:
        -----------
        wait (bool) : ask the helm client to wait until resources are created
            before returning.

        debug (bool) : ask the helm client for verbose output

        """
        for repo_name, repo_url in self.repos.items():
            self._run(cmd=['repo', 'add', repo_name, repo_url])

        cmd = ['upgrade']
        cmd += [self.name, self.chart]

        cmd += ['--install', '--create-namespace']

        if wait:
            cmd += ['--wait']
        if debug:
            cmd += ['--debug']

        if len(self.set):
            # turn the set dict into '--set a=A,b=B,c=C'
            cmd += ['--set', ','.join([f"{name}={value}" for (name, value) in self.values.items()])]

        if len(self.values):
            # turn the values into a file, and add it to the command
            with open(self.file, 'w') as val_file:
                yaml.dump(self.values, val_file)

            cmd += ['--values', self.file]

        try:
            self._run(cmd=cmd)
        except Exception as err:
            raise RuntimeError("Helm failed to install the relese") from err

    # -all is the used command flag, so the var name makes sense
    # pylint: disable=redefined-builtin
    def list(self, all: bool = False, failed: bool = False,
             deployed: bool = False, pending: bool = False):
        """List all releases.

        This is not instance specific but still useful.

        """
        cmd = ['list', '--output=yaml']

        if all:
            cmd += ['--all']
        elif deployed:
            cmd += ['--deployed']
        elif failed:
            cmd += ['--failed']
        elif pending:
            cmd += ['--pending']

        list_str = self._run(cmd=cmd, return_output=True)
        if list_str:
            return yaml.safe_load(list_str)

        return []

    def destroy(self, debug: bool = False):
        """Remove an installed helm release.

        Parameters:
        -----------
        debug (bool) : ask the helm client for verbose output

        """
        cmd = ['uninstall', self.name]

        if debug:
            cmd += ['--debug']

        self._run(cmd=cmd)

    def test(self):
        """Test an installed helm release.

        This runs the helm client test command.

        """
        self._run(cmd=['test', self.name])

    def status(self):
        """Get status of the installed helm release."""
        return HelmReleaseStatus(yaml.safe_load(
            self._run(cmd=['status', self.name, '--output', 'yaml'], return_output=True)))

    def _run(self, cmd: List[str], return_output: bool = False):
        """Run a helm v3 command."""
        cmd = [self.bin, f'--kubeconfig={self.kubeconfig}', f'--namespace={self.namespace}'] + cmd

        # this syntax makes it easier to read
        # pylint: disable=no-else-return
        if return_output:
            logger.debug("running launchpad command with output capture: %s", " ".join(cmd))
            exec = subprocess.run(cmd, cwd=self.working_dir, shell=False, check=True,
                                  stdout=subprocess.PIPE)
            exec.check_returncode()
            return exec.stdout.decode('utf-8')
        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            exec = subprocess.run(cmd, cwd=self.working_dir, check=True, text=True)
            exec.check_returncode()
            return exec


class Status(Enum):
    """A Helm Status enum."""

    DEPLOYED = 'deployed'
    """ Helm Release has been deployed """


# This is effectively a struct for interpreting the release status information
# pylint: disable=too-few-public-methods
class HelmReleaseStatus:
    """Interpreted helm release status.

    Used to formalize the response object from a status request

    """

    def __init__(self, status_reponse: dict):
        """Interpret status from a status response."""
        self.name = status_reponse['name']
        self.version = status_reponse['version']
        self.namespace = status_reponse['namespace']

        # self.first_deployed = datetime.strptime(status_reponse['info']['first_deployed'],
        #                                         '%Y-%m-%dT%H:%M:%S.%fZ')
        # self.last_deployed = datetime.strptime(status_reponse['info']['last_deployed'],
        #                                         '%Y-%m-%dT%H:%M:%S.%fZ')

        self.deleted = status_reponse['info']['deleted']
        self.description = status_reponse['info']['description']
        self.status = Status(status_reponse['info']['status'])
        self.notes = status_reponse['info']['notes']

        self.config = status_reponse['config']

        self.manifect = status_reponse['manifest']



"""

Dummy metta provisioner plugin

A provisioner which doesn't do anything, but can still be configured to produce
various clients and outputs.  The provisioner has no significant requirements
and no impact.  Provisioning can be repeated or interupted without impact.

The dummy provisioner is entirely config based.  Use the prepare() method to
indicate an appropriate config source and the provisioner will do the rest. As
long as you match its config convention it can take care of itself.

"""

import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.provisioner import ProvisionerBase

logger = logging.getLogger('metta.contrib.dummy.provisioner')


class DummyProvisionerPlugin(ProvisionerBase):
    """ Dummy provisioner class """

    def __init__(self, environment: Environment, instance_id: str,
                 fixtures: Dict[str, Dict[str, Any]] = None):
        """ Run the super constructor but also set class properties """
        super().__init__(environment, instance_id)

        if fixtures is not None:
            fixtures = environment.add_fixtures_from_dict(plugin_list=fixtures)
        else:
            fixtures = Fixtures()
        self.fixtures: Fixtures = fixtures
        """ All fixtures added to this dummy plugin. """

    def apply(self):
        """ pretend to bring a cluster up """
        logger.info("%s:execute: apply()", self.instance_id)

    def prepare(self):
        """ pretend to prepare the cluster """
        logger.info("%s:execute: apply()", self.instance_id)

    def destroy(self):
        """ pretend to brind a cluster down """
        logger.info("%s:execute: apply()", self.instance_id)



"""

Metta terraform CLI plugin.

Provides functionality to inspect terraform configuration and execute
terraform operations, as well as check out terraform outputs.

"""
import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .provisioner import METTA_TERRAFORM_PROVISIONER_PLUGIN_ID

logger = logging.getLogger('metta.cli.terraform')

METTA_TERRAFORM_CLI_PLUGIN_ID = 'metta_terraform'
""" cli plugin_id for the info plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class TerraformCliPlugin(CliBase):
    """Fire command/group generator for terraform commands."""

    def fire(self):
        """Return a dict of commands.

        Don't return any commands if there is no provisioner plugin available

        """
        if self.environment.fixtures.get(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                                         plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
                                         exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'terraform': TerraformGroup(self.environment)
                }
            }

        return {}


class TerraformGroup():
    """Base Fire command group for terraform cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """Pick a matching terraform provisioner."""
        if instance_id:
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
                instance_id=instance_id)

        # Get the highest priority provisioner
        return self.environment.fixtures.get(
            plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
            plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID)

    def info(self, provisioner: str = '', deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            if hasattr(fixture.plugin, 'get_fixtures'):
                fixtures = {}
                for sub_fixture in fixture.plugin.get_fixtures().to_list():
                    fixture_info = {
                        'fixture': {
                            'plugin_type': sub_fixture.plugin_type,
                            'plugin_id': sub_fixture.plugin_id,
                            'instance_id': sub_fixture.instance_id,
                            'priority': sub_fixture.priority,
                        }
                    }
                    if hasattr(sub_fixture.plugin, 'info'):
                        plugin_info = sub_fixture.plugin.info()
                        if isinstance(plugin_info, dict):
                            fixture_info.update(plugin_info)
                    fixtures[sub_fixture.instance_id] = fixture_info
                info['fixtures'] = fixtures

        return cli_output(info)

    def fixtures(self, provisioner: str = ''):
        """List all fixtures for this provisioner."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')
        fixture_list = [{
            'plugin_type': fixture.plugin_type,
            'plugin_id': fixture.plugin_id,
            'instance_id': fixture.instance_id,
            'priority': fixture.priority,
        } for fixture in provisioner_plugin.get_fixtures().to_list()]

        cli_output(fixture_list)

    def prepare(self, provisioner: str = ''):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()

    def apply(self, provisioner: str = ''):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply()

    def check(self, provisioner: str = ''):
        """Run provisioner check."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.check()

    def destroy(self, provisioner: str = ''):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy()



"""

METTA CLI setuptools entrypoint

Keep the entrypint minimal, and rely on .base for real work.

"""

import logging

import fire

from .root import Root

logger = logging.getLogger('metta.cli.entrypoint')


def main():
    """ Main entrypoint """
    fire.Fire(Root)


if __name__ == '__main__':
    main()



"""

Metta Mirantis.

Common components and configuration which are nice to use
when using metta with Mirantis products.

THese are entirely optional, but it makes a lot of
operations easier.

"""

from typing import List, Dict

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .common import add_common_config
from .presets import add_preset_config

from .mke_client import MKEAPIClientPlugin, MKEAPICliPlugin, METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID
from .msr_client import MSRAPIClientPlugin, MSRAPICliPlugin, METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID

# ----- Plugin factories -----


# pylint: disable=too-many-arguments
@Factory(plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
def metta_mirantis_plugin_factory_client_mke(
        environment: Environment,
        instance_id: str = "",
        accesspoint: str = '',
        username: str = '',
        password: str = '',
        hosts: List[Dict] = None):
    """Create a Mirantis MKE API Client."""
    return MKEAPIClientPlugin(environment, instance_id, accesspoint=accesspoint,
                              username=username, password=password, hosts=hosts)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
def metta_terraform_factory_cli_mke(
        environment: Environment, instance_id: str = ''):
    """Create an MKE cli plugin."""
    return MKEAPICliPlugin(environment, instance_id)


# pylint: disable=too-many-arguments
@Factory(plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)
def metta_mirantis_plugin_factory_client_msr(
        environment: Environment,
        instance_id: str = '',
        accesspoint: str = '',
        username: str = '',
        password: str = '',
        hosts: List[Dict] = None):
    """Create a Mirantis MSR API Client."""
    return MSRAPIClientPlugin(environment, instance_id, accesspoint=accesspoint,
                              username=username, password=password, hosts=hosts)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)
def metta_terraform_factory_cli_msr(
        environment: Environment, instance_id: str = ''):
    """Create an MSR cli plugin."""
    return MSRAPICliPlugin(environment, instance_id)


# ----- METTA bootstraps that we will use on config objects -----


def bootstrap_common(environment: Environment):
    """Metta configerus bootstrap.

    Add some Mirantis specific config options for presets

    Additional config sources may be added based on config.load("metta")
    which will try to add config for metta `presets`
    This gives access to "variation", "cluster", "platform" and "release"
    presets.  You can dig deeper, but it might implicitly make sense if you look
    at the config folder of this module.

    @see .common.add_common_config() for details

    """
    add_common_config(environment)


def bootstrap_presets(environment: Environment):
    """Metta configerus bootstrap.

    Add some Mirantis specific config options for presets

    Additional config sources may be added based on config.load("metta")
    which will try to add config for metta `presets`
    This gives access to "variation", "cluster", "platform" and "release"
    presets.  You can dig deeper, but it might implicitly make sense if you look
    at the config folder of this module.

    @see .presets.add_preset_config() for details

    """
    add_preset_config(environment)



"""

Docker workload plugin to run docker run commands, given docker host config.

This type of docker workload is simple to use but perhaps not
very feature rich.

"""
from typing import Any
import logging

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

logger = logging.getLogger('metta.contrib.docker.workload.run')

DOCKER_RUN_WORKLOAD_CONFIG_LABEL = 'docker'
""" Configerus label for retrieving docker run workloads """
DOCKER_RUN_WORKLOAD_CONFIG_BASE = 'workload.run'
""" Configerus get base for retrieving the default run workload """


class DockerPyRunWorkloadPlugin(WorkloadBase):
    """Docker Run workload class for the DockerPy."""

    def __init__(
            self,
            environment: Environment,
            instance_id: str,
            label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
            base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE):
        """Run the super constructor but also set class properties.

        Parameters:
        -----------
        outputs (Dict[Dict]) : pass in a dictionary which defines outputs that
            should be returned

        clients (Dict[Dict]) : pass in a dictionary which defines which clients
            should be requested when working on a provisioner

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        logger.info("Preparing Docker run setting")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a docker client plugin.

        """
        client = fixtures.get_plugin(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id='metta_docker')

        run_config = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the docker run label """

        run = run_config.get([self.config_base, 'run'])

        return DockerRunWorkloadInstance(client, run)

    def info(self):
        """Return dict data about this plugin for introspection."""
        run_config = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the docker run label """

        return {
            'workload': {
                'run': {
                    'run': run_config.get([self.config_base, 'run'])
                },
                'required_fixtures': {
                    'docker': {
                        'plugin_type': METTA_PLUGIN_TYPE_CLIENT,
                        'plugin_id': 'metta_docker'
                    }
                }
            }
        }


# This is a small workload instance, but it does its job
# pylint: disable=too-few-public-methods
class DockerRunWorkloadInstance(WorkloadInstanceBase):
    """ A workload instance for a docker run """

    def __init__(self, client, run):
        """Inject docker client, and run definition into workload instance."""
        self.client = client
        self.run = run

    def apply(self):
        """Run the workload.

        @NOTE Needs a docker client fixture to run.  Use .set_fixtures() first

        """
        assert 'image' in self.run, "Run command had no image"

        self.run = self.client.containers.run(**self.run)
        return self.run

    # @TODO write the docker container rm
    def destroy(self):
        """Destroy a created docker run."""



"""

CLI Plugins base.

@TODO move to the metta_cli package

Define a base class and some constants for the CLI plugins.  The CLI plugins provide functionality
for the metta cli tool, which is provided by the metta_cli package.  This is here because the plugin
types are defined as an enum in this package, and so the base class is here too.

"""
import logging

from mirantis.testing.metta import environment_names, get_environment, discover, new_environment
from mirantis.testing.metta.plugin import Factory

from .base import METTA_PLUGIN_TYPE_CLI


logger = logging.getLogger('metta.cli.root')


# This gets used to collect public properties dynamically.
# pylint: disable=too-few-public-methods
class Root:
    """The Metta CLI program.

    This CLI let's you interact with a metta environment for the purpose of
    introspection, debugging and manual interaction.

    All groups and commands come from metta CLI plugins.

    """

    def __init__(self, environment: str = '', state: str = ''):
        """Configure initial root object.

        Parameters:
        -----------
        environment (str) : Environment name in case you want to switch to an
            alternate environment.

        state (str) : Environment state to consider active.  Will throw an error
            if the state doesn't exist for the selected environment.

        """
        # Try to make an environment using the core discover approach.  This looks for a metta.yml
        # file and uses it as a config source, to load all other configuration.
        discover()

        try:
            if environment == '':
                environment = environment_names()[0]
        except (KeyError, IndexError):
            logger.warning("No environment object has been defined (making one now.) "
                           "Are you in a project folder?")
            environment = 'empty'
            new_environment(environment)

        try:
            self._environment = get_environment(environment)

            if state:
                self._environment.set_state(state)
        except KeyError as err:
            raise ValueError(f"Could not load environment '{environment}', not found") from err

        # collect any comands from all discovered cli plugins
        self._collect_commands()

    def _collect_commands(self):
        """Collect commands from all cli plugins.

        Create an instance of any registered cli plugin.
        From the plugin, collect the commands and add each command to this
        object directly, so that Fire can see them.

        """
        for plugin_id in Factory.registry[METTA_PLUGIN_TYPE_CLI]:
            plugin_dict = {
                'plugin_type': METTA_PLUGIN_TYPE_CLI,
                'plugin_id': plugin_id
            }

            fixture = self._environment.add_fixture_from_dict(plugin_dict=plugin_dict)

            logger.info("loading cli plugin: %s", fixture.plugin_id)
            plugin = fixture.plugin

            if not hasattr(plugin, 'fire'):
                continue

            try:
                commands = plugin.fire()
            except TypeError as err:
                raise NotImplementedError(f"Plugin {plugin.plugin_id} did not properly implement "
                                          f"the fire(fixtures) method") from err

            if not isinstance(commands, dict):
                raise ValueError(f"Plugin returned invalid commands : {commands}")

            # Attach any found commands/groups to this object
            for (command_name, command) in commands.items():
                logger.debug("adding cli plugin command/group: %s->%s",
                             fixture.plugin_id,
                             command_name)

                # if the command name already exists and is a dict then
                # maybe we should merge them
                if hasattr(self, command_name) and isinstance(
                        getattr(self, command_name), dict) and isinstance(command, dict):
                    getattr(self, command_name).update(command)
                    continue

                setattr(self, command_name, command)



"""

Fixture management code.

A Fixture is a plugin isntance wrapper that keeps metadata about the plugin
instance. A Fixtures object is a collection of plugin instances kept as an
ordered, managed set.

"""
import logging
from typing import List, Iterator

# pylint: disable=W0511
# TODO move these to this file as METTA_FIXTURE_KEY_XXXXX
from .plugin import (METTA_PLUGIN_CONFIG_KEY_PLUGINTYPE, METTA_PLUGIN_CONFIG_KEY_PLUGINID,
                     METTA_PLUGIN_CONFIG_KEY_INSTANCEID, METTA_PLUGIN_CONFIG_KEY_PRIORITY)

logger = logging.getLogger('metta.fixtures')

METTA_FIXTURES_CONFIG_FIXTURES_LABEL = 'fixtures'
""" A centralized configerus load label for multiple fixtures """
METTA_FIXTURES_CONFIG_FIXTURE_KEY = 'fixture'
""" Config .get() key for a single fixture """

METTA_FIXTURE_VALIDATION_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        METTA_PLUGIN_CONFIG_KEY_PLUGINID: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_INSTANCEID: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_PLUGINTYPE: {'type': 'string'},
        METTA_PLUGIN_CONFIG_KEY_PRIORITY: {
            'type': 'integer', 'minimum': 1, 'maximum': 100}
    },
    'required': [METTA_PLUGIN_CONFIG_KEY_PLUGINID]
}
""" json schema validation definition for a plugin """


# pylint: disable=too-few-public-methods
class Fixture:
    """A plugin wrapper struct that keep metadata about the plugin in a set.

    pylint: R0903 ; if we replace this with a Dict then we have to define the
        key values as a constant set.  it would be cumbersome to import such
        things all over the place.

    """

    # pylint: disable=too-many-arguments
    def __init__(self, plugin: object, plugin_type: str, plugin_id: str, instance_id: str,
                 priority: int):
        """Initialize struct contents.

        Parameters:
        -----------
        plugin : the fixture plugin instance

        Filtering parameters:

        plugin_type (str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        """
        self.plugin_type = plugin_type
        self.plugin_id = plugin_id
        self.instance_id = instance_id
        self.priority = priority
        self.plugin = plugin

    def __eq__(self, other):
        """Compare to another fixture.

        Parameters:
        -----------
        other (Fixture) : fixture to compare this fixture against.

        Returns:
        --------
        Boolean True if the other fixture matches this one in metadata,
        otherwise False.

        """
        if not isinstance(other, Fixture):
            return False

        return (self.plugin_type == other.plugin_type
                and self.plugin_id == other.plugin_id
                and self.instance_id == other.instance_id)


class Fixtures:
    """A managed set of Fixture objects.

    A set of plugins that can be added in an arbitrary order but retrieved
    using filters and sorting.  The plugins are wrapped as Fixtures, which are
    used for the sorting and filtering.

    To use this, create an instance, and then add fixtures using either new()
    or add().  You can also combine two collections using merge().
    To get fixtures out, either iterator across the Fixtures object, or use
    get() to retrieve a single matching item.  use filter() to produce a new
    Fixtures object as a reduced set.

    """

    def __init__(self):
        """Create initial empty fixtures list."""
        self._fixtures: List[Fixture] = []
        """ object List of fixtures. """

    def __len__(self) -> int:
        """Return how many plugin instances we have.

        Returns:
        --------
        Integer length of the collection of Fixture objects.

        """
        return len(self._fixtures)

    def __iter__(self) -> Iterator[Fixture]:
        """Create an iterator for the fixtures object.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Fixture objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return iter(self.to_list())

    def __reversed__(self) -> Iterator[Fixture]:
        """Create a reversed iterator for the fixtures object.

        @TODO switch to just using a generator?

        Returns:
        --------
        An Iterator of Fixture objects

        """
        # Iterate across the to_list() set, as it is sorted.
        return reversed(self.to_list())

    def merge(self, merge_from: 'Fixtures'):
        """Merge fixture instances from another Fixtures object into this one.

        Parameters:
        -----------
        Merge_from (Fixtures) : fixture instance source

        Raises:
        -------
        May raise a KeyError if there a matching plugin is already in the
        Fixtures object.

        (plugin_type/plugin_id/instance_id)

        """
        # We use the add_fixture method to centralize the logic for adding fixtures to one function
        for fixture in merge_from:
            self.add(fixture)

    # pylint: disable=too-many-arguments
    def new(self, plugin: object, plugin_type: str, plugin_id: str,
            instance_id: str, priority: int) -> Fixture:
        """Add a new fixture by providing the plugin instance and the metadata.

        Create a new Fixture from the passed arguments and add it to the
        Fixtures set.

        Parameters:
        -----------
        plugin : the fixture plugin instance

        Filtering parameters:

        plugin_type (str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier
        priority (int) : plugin priority

        """
        return self.add(Fixture(plugin_type=plugin_type, plugin_id=plugin_id,
                                instance_id=instance_id, priority=priority, plugin=plugin))

    def add(self, fixture: Fixture, allow_overwrite: bool = True) -> Fixture:
        """Add an existing fixture.

        If a matching

        Parameters:
        -----------
        fixture (Fixture) : existing fixture to add

        allow_overwrite (bool) : If True, then this fixture can replace a
            matching plugin with the same metadata.

        """
        if self.get(plugin_type=fixture.plugin_type, plugin_id=fixture.plugin_id,
                    instance_id=fixture.instance_id, exception_if_missing=False):

            if not allow_overwrite:
                raise KeyError("Fixture index already exists:"
                               f"[plugin_type:{fixture.plugin_type}][plugin_id:{fixture.plugin_id}]"
                               f"[instance_id:{fixture.instance_id}]")

            for i in range(len(self._fixtures)):
                if self._fixtures[i] == fixture:
                    self._fixtures[i] = fixture
                    return fixture

        self._fixtures.append(fixture)

        return fixture

    def get(self, plugin_type: str = '', plugin_id: str = '', instance_id: str = '',
            exception_if_missing: bool = True) -> Fixture:
        """Retrieve the first matching fixture object based on filters and priority.

        Parameters:
        -----------
        Filtering parameters:

        plugin_type (str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was
        found.

        """
        filtered = self.filter(
            plugin_type=plugin_type,
            plugin_id=plugin_id,
            instance_id=instance_id)

        if len(filtered) > 0:
            return filtered.to_list()[0]
        if exception_if_missing:
            raise KeyError("Could not find any matching fixture instances "
                           f"[plugin_type:{plugin_type}][plugin_id:{plugin_id}]"
                           f"[instance_id:{instance_id}]")
        # filtered list was empty, and we were not directed to raise an exception for that.
        return None

    def get_plugin(self, plugin_type: str = '', plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> object:
        """Retrieve the first matching plugin  based on filters and priority.

        Parameters:
        -----------
        Filtering parameters:

        plugin_type (str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        The highest priority matched Fixture fixture.
        If now fixtures matched, and exception_if_missing is False, then None

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was
        found. This error is actually raise in self.get()

        """
        fixture = self.get(
            plugin_type=plugin_type,
            plugin_id=plugin_id,
            instance_id=instance_id,
            exception_if_missing=exception_if_missing)

        if fixture is None:
            return None
        return fixture.plugin

    def filter(self, plugin_type: str = '', plugin_id: str = '', instance_id: str = '',
               exception_if_missing: bool = False) -> 'Fixtures':
        """Filter the fixture instances.

        Parameters:
        -----------
        Filtering parameters:

        plugin_type (str) : Type of plugin
        plugin_id (str) : registry plugin_id
        instance_id (str) : plugin instance identifier

        Returns:
        --------
        A new Fixtures object with only matching Fixture objects.  It could
        contain all of the items. If no filters were passed in, or it could
        be empty if no matches were found and the passed exception_is_missing
        variable is True

        Raises:
        -------
        KeyError if exception_if_missing is True and no matching fixture was found

        """
        filtered = Fixtures()
        for fixture in self._fixtures:
            # Could have one-lined this into a lambda but it would be
            # unreadable

            if plugin_type and not fixture.plugin_type == plugin_type:
                continue
            if plugin_id and not fixture.plugin_id == plugin_id:
                continue
            if instance_id and not fixture.instance_id == instance_id:
                continue

            filtered.add(fixture)

        if exception_if_missing and len(filtered) == 0:
            raise KeyError(f"Filter found matches [{plugin_type}][{plugin_id}][{instance_id}]")
        return filtered

    def to_list(self) -> List[Fixture]:
        """Order the fixtures based on priority.

        This is usefull if you need a list, but you can rely on the
        Fixtures.__iter__ for most scenarios, and just iterate across the
        Fixtures object itself.

        This method is relied on internally.

        Returns:
        --------
        A List[Fixture] of the contained fixtures sorted using their priority.

        """
        return sorted(self._fixtures, key=lambda i: 1 / i.priority if i.priority > 0 else 0)



"""

CLI client for running launchpad commands as an injectable class.

All of the metta plugijn functionality hands off to this module for actually
executing commands.

"""
import os
import logging
import json
import datetime
import subprocess
import shutil
from typing import Dict, List

import yaml

logger = logging.getLogger('metta_launchpad:launchpad')

METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT = './launchpad.yml'
""" Launchpad config configuration file key """

METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT = '.'
""" Launchpad Client default working dir """
METTA_LAUNCHPADCLIENT_BIN_PATH = 'launchpad'
""" Launchpad bin exec for the subprocess """

METTA_USER_LAUNCHPAD_CLUSTER_PATH = os.path.expanduser(
    os.path.join('~', '.mirantis-launchpad', 'cluster'))
""" the str path to where launchpad keeps its user config """
METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH = 'bundle'
""" str path to user bundle config can be found when it is downloaded """
METTA_USER_LAUNCHPAD_BUNDLE_META_FILE = 'meta.json'
""" str filename for the meta file in the client bundle path """

LAUNCHPAD_CLIENT_BUNDLE_RETRY_COUNT_DEFAULT = 2
""" default value for how many times we should retry client bundle download """


# yeah, this should probably get reduced.  Maybe combine the bools into a map
# pylint: disable=too-many-instance-attributes
class LaunchpadClient:
    """Shell client for interacting with the launchpad bin."""

    # this is what it takes to configure launchpad usage
    # pylint: disable=too-many-arguments
    def __init__(self, config_file: str = METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT,
                 working_dir: str = METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT,
                 cluster_name_override: str = '', accept_license: bool = False,
                 disable_telemetry: bool = False, disable_upgrade_check: bool = True,
                 debug: bool = False):
        """Initialize LaunchpadClient object.

        Parameters:
        -----------
        config_file (str) : path to launchpad config file, typically
            launchpad.yml

        working_dir (str) : full config file path.
            this typically plays a role in interpreting file paths from the
            config for things like ssh keys.  The client will use that path for
            python subprocess execution

        cluster_name_override (str) : string name for the cluster which should
            be used instead of trying to read a value from config.  This can
            can help in some scenarios where config cannot be read outside of
            the client, and so cluster name is hard to discover.

        accept_license (bool) : passed to the launchpad client to tell it to
            accept the license on first use.

        disable_telemetry (bool) : passed to the launchpad client to tell it to
            disable telemetry on actions taken.

        disable_upgrade_check (bool) : passed to the launchpad client to tell
            it to disable checking to see if a new client version is available.

        debug (bool) : passed to the launchpad client to tell it to enable
            verbose debugging output.

        """
        self.config_file: str = config_file
        """ Path to config file """
        self.working_dir: str = working_dir
        """ Python subprocess working dir to execute launchpad in.
            This may be relevant in cases where ssh keys have relative path """

        self.cluster_name_override: str = str(cluster_name_override).strip()
        """ if not empty this will be used as a cluster name instead of taking
            it from the yaml file """

        self.bin: str = METTA_LAUNCHPADCLIENT_BIN_PATH
        """ shell execution target for launchpad """

        self.client_bundle_retry_count: int = int(LAUNCHPAD_CLIENT_BUNDLE_RETRY_COUNT_DEFAULT)
        """ how many times to rety a client bundle download """

        self.disable_telemetry: bool = disable_telemetry
        self.accept_license: bool = accept_license
        self.disable_upgrade_check: bool = disable_upgrade_check

        self.debug: bool = debug
        """ should launchpad be run with verbose output enabled ? """

    def version(self):
        """Output launchpad client version."""
        self._run(['version'])

    def apply(self, debug: bool = False):
        """Install using the launchpad client."""
        self._run(['apply'], debug=debug)

    def exec(self, host_index: int, cmds: List[str]):
        """Execute a command on a host index."""
        client_config = self.describe_config()
        hosts = client_config['spec']['hosts']
        host = hosts[host_index]
        if 'ssh' in host:
            target = host['ssh']['address']
        if 'winrm' in host:
            target = host['winrm']['address']

        args = ['exec', '--target', target]
        args.extend(cmds)
        self._run(args)

    def exec_interactive(self, host_index: int, cmds: List[str]):
        """Execute a command on a host index."""
        client_config = self.describe_config()
        hosts = client_config['spec']['hosts']
        host = hosts[host_index]
        if 'ssh' in host:
            target = host['ssh']['address']
        if 'winrm' in host:
            target = host['winrm']['address']
        args = ['exec', '--interactive', '--target', target]
        args.extend(cmds)
        self._run(args)

    def reset(self, quick: bool = False):
        """Uninstall using the launchpad client."""
        if os.path.isfile(self.config_file):
            if not quick:
                self._run(['reset', '--force'])
            self.rm_client_bundles()

            try:
                os.remove(self.config_file)
            except OSError:
                pass

    def register(self, name: str, email: str, company: str):
        """Uninstall using the launchpad client."""
        return self._run(['register', '--name', name,
                          '--email', email, '--company', company])

    def describe_config(self):
        """Return the launchpad config report as umarshalled yaml."""
        return yaml.safe_load(
            self._run(['describe', 'config'], return_output=True))

    def describe(self, report: str):
        """Output one of the launchpad reports."""
        self._run(['describe', report])

    def bundle_users(self):
        """List bundle users which have been downloaded."""
        return self._mke_client_downloaded_bundle_user_paths().keys()

    def bundle(self, user: str, reload: bool = False):
        """Retrieve a client bundle and return the metadata as a dict."""
        client_bundle_path = self._mke_client_bundle_path(user)
        client_bundle_meta_file = os.path.join(
            client_bundle_path, METTA_USER_LAUNCHPAD_BUNDLE_META_FILE)

        if reload or not os.path.isfile(client_bundle_meta_file):

            # @NOTE currently client bundle downloads are flaky.  They fail about 1/5 times
            #    with unclear TLS issues.  Because the failures are intermittent, we should
            #    just try again

            for i in range(1, self.client_bundle_retry_count):
                try:
                    self._run(["client-config", user])
                    break
                except subprocess.CalledProcessError as err:
                    logger.warning("Attempt %s to download bundle failed.  Assuming flaky "
                                   "behaviour and trying again : %s", i, err)

            else:
                raise RuntimeError("Numerous attempts to download the client bundle failed.")

        data = {}
        """ Will hold data pulled from the client meta data file """
        try:
            with open(client_bundle_meta_file) as json_file:
                data = json.load(json_file)

                # helm complains if this file has loose permissions
                client_bundle_kubeconfig_file = os.path.join(
                    client_bundle_path, 'kube.yml')
                os.chmod(client_bundle_kubeconfig_file, 0o600)
        except FileNotFoundError as err:
            raise ValueError(f"failed to open the launchpad client bundle meta "
                             f"file : {client_bundle_meta_file}") from err

        # Not sure why this isn't in there:
        data['Endpoints']['kubernetes']['kubeconfig'] = client_bundle_kubeconfig_file
        # add some stuff that a client bundle always has
        data['path'] = client_bundle_path
        data['modified'] = datetime.datetime.fromtimestamp(
            os.path.getmtime(client_bundle_meta_file)).strftime("%Y-%m-%d %H:%M:%S")
        # this stuff should already be in the bundle, but it isn't
        data['tls_paths'] = {
            'docker': os.path.join(client_bundle_path, 'tls', 'docker'),
            'kubernetes': os.path.join(client_bundle_path, 'tls', 'kubernetes'),
        }

        return data

    def rm_client_bundles(self):
        """Remove any downloaded client bundles."""
        try:
            base = os.path.join(
                METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                self._cluster_name(),
                METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH)

            if os.path.isdir(base):
                shutil.rmtree(base)

        except subprocess.CalledProcessError:
            # most likely we could't determine a cluster_name, because we don't
            # have a config file.
            return

    def _mke_client_bundle_root(self):
        """Root path to the launchpad user conf."""
        return os.path.join(
            METTA_USER_LAUNCHPAD_CLUSTER_PATH, self._cluster_name())

    def _mke_client_bundle_path(self, user: str):
        """Find the path to a client bundle for a user."""
        return os.path.join(METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                            self._cluster_name(), METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH, user)

    def _mke_client_downloaded_bundle_user_paths(self) -> Dict[str, str]:
        """Return a map of user names to downloaded bundle paths."""
        try:
            base = os.path.join(
                METTA_USER_LAUNCHPAD_CLUSTER_PATH,
                self._cluster_name(),
                METTA_USER_LAUNCHPAD_BUNDLE_SUBPATH)
            return {userdir: os.path.join(base, userdir) for userdir in os.listdir(
                base) if os.path.isdir(os.path.join(base, userdir))}
        except (ValueError, FileNotFoundError):
            logger.debug("Could not get user bundles names as there are"
                         " no launchpad targets to check against.")
            return {}

    def _cluster_name(self):
        """Get the cluster name from the config file.

        @TODO we should cache it, but then do we need to invalidate cache?

        """
        if self.cluster_name_override:
            return str(self.cluster_name_override)

        try:
            with open(self.config_file) as config_file_object:
                config_data = yaml.safe_load(config_file_object)
                """ keep a parsed copy of the launchpad file """
        except FileNotFoundError as err:
            raise ValueError("Launchpad yaml file could not be opened"
                             f": {self.config_file}") from err
        except Exception as err:
            raise ValueError("Launchpad yaml file had unexpected contents:"
                             f" {self.config_file}") from err

        if not isinstance(config_data, dict):
            raise ValueError(f"Launchpad yaml file had unexpected contents: {self.config_file}")

        try:
            return str(config_data['metadata']['name'])
        except KeyError as err:
            raise ValueError('Launchpad yaml file did not container a cluster name') from err

    def _run(self, args: List[str], return_output=False, debug: bool = False):
        """Run a launchpad command.

        Parameters:
        -----------
        args (List[str]) : arguments to pass to the launchpad bin using
            subprocess

        return_output (bool) : this call should capture the exec output and
            return it instead of sending it to stdout

        debug (bool) : override class debug value if True

        """
        # if the command passed uses a config file, add the flag for it
        if not args[0] in ['help', 'version']:
            args = [args[0]] + ['-c', self.config_file] + args[1:]

        cmd = [self.bin]

        cmd += [args[0]]

        if debug or self.debug:
            cmd += ['--debug']
        if self.disable_telemetry:
            cmd += ['--disable-telemetry']
        if self.accept_license:
            cmd += ['--accept-license']
        if self.disable_upgrade_check:
            cmd += ['--disable-upgrade-check']

        if len(args) > 1:
            cmd += args[1:]

        # makes it more readable
        # pylint: disable=no-else-return
        if return_output:
            logger.debug("running launchpad command with output capture: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, shell=False, check=True,
                                 stdout=subprocess.PIPE)
            res.check_returncode()
            return res.stdout.decode('utf-8')

        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            res = subprocess.run(cmd, cwd=self.working_dir, check=True, text=True)
            res.check_returncode()
            return res



"""

Metta CLI : Provisioner commands.

Various commands that allow introspection of provisioner plugins/fixtures and
their contents.

"""
import logging


from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER

from .base import CliBase, cli_output

logger = logging.getLogger('metta.cli.provisioner')


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class ProvisionerCliPlugin(CliBase):
    """Fire command/group generator for provisioner commands."""

    def fire(self):
        """Return a dict of commands."""
        return {
            'provisioner': ProvisionerGroup(self.environment)
        }


class ProvisionerGroup():
    """Base Fire command group for provisioner commands."""

    def __init__(self, environment: Environment):
        """Create CLI command group."""
        self.environment = environment

    def list(self, raw: bool = False):
        """List all provisioners."""
        provisioner_list = [
            fixture.plugin.instance_id for fixture in self.environment.fixtures.filter(
                plugin_type=METTA_PLUGIN_TYPE_PROVISIONER)]

        if raw:
            return list
        return cli_output(provisioner_list)

    def _select_provisioner(self, instance_id: str = ''):
        """Pick a matching provisioner."""
        if instance_id:
            return self.environment.fixtures.get(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER,
                                                 instance_id=instance_id)
        # Get the highest priority provisioner
        return self.environment.fixtures.get(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER)

    def info(self, provisioner: str = '', deep: bool = True):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            if hasattr(fixture.plugin, 'get_fixtures'):
                fixtures = {}
                for sub_fixture in fixture.plugin.get_fixtures().to_list():
                    fixture_info = {
                        'fixture': {
                            'plugin_type': sub_fixture.plugin_type,
                            'plugin_id': sub_fixture.plugin_id,
                            'instance_id': sub_fixture.instance_id,
                            'priority': sub_fixture.priority,
                        }
                    }
                    if hasattr(sub_fixture.plugin, 'info'):
                        plugin_info = sub_fixture.plugin.info()
                        if isinstance(plugin_info, dict):
                            fixture_info.update(plugin_info)
                    fixtures[sub_fixture.instance_id] = fixture_info
                info['fixtures'] = fixtures

        return cli_output(info)

    def fixtures(self, provisioner: str = '', deep: bool = False):
        """List all fixtures for this provisioner."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner_plugin, 'get_fixtures'):
            raise ValueError('This provisioner does not keep fixtures.')

        fixture_list = {}
        for fixture in provisioner_plugin.get_fixtures().to_list():

            info = {
                'fixture': {
                    'plugin_type': fixture.plugin_type,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }

            if deep:
                if hasattr(fixture.plugin, 'info'):
                    info.update(fixture.plugin.info())

                if hasattr(fixture.plugin, 'get_fixtures'):
                    fixtures = {}
                    for sub_fixture in fixture.plugin.get_fixtures().to_list():
                        fixture_info = {
                            'fixture': {
                                'plugin_type': sub_fixture.plugin_type,
                                'plugin_id': sub_fixture.plugin_id,
                                'instance_id': sub_fixture.instance_id,
                                'priority': sub_fixture.priority,
                            }
                        }
                        if hasattr(sub_fixture.plugin, 'info'):
                            fixture_info.update(sub_fixture.plugin.info())

                        fixtures[sub_fixture.instance_id] = fixture_info

                    info['fixtures'] = fixtures

            fixture_list[fixture.instance_id] = info

        return cli_output(fixture_list)

    def prepare(self, provisioner: str = ''):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()

    def apply(self, provisioner: str = ''):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply()

    def destroy(self, provisioner: str = ''):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy()



"""

PLUGIN: Provisioner.

Provisioning plugins meet a common interface standard for managing a testing
system, where the provisioning keywords relate to creating and tearing down the
system.

"""

import logging

logger = logging.getLogger('metta.provisioner')

METTA_PLUGIN_TYPE_PROVISIONER = 'provisioner'
""" metta plugin type idnetifier for client plugins """

METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL = 'provisioners'
""" A centralized configerus load labe for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL = 'provisioner'
""" A centralized configerus load label a provisioner """
METTA_PROVISIONER_CONFIG_PROVISIONERS_KEY = 'provisioners'
""" A centralized configerus key for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_KEY = 'provisioner'
""" A centralized configerus key for one provisioner """


class ProvisionerBase:
    """Base Provisioner plugin class."""

    def prepare(self):
        """Prepare the provisioner to apply resources.

        Initial Provisioner plugin is expected to be of very low cost until
        prepare() is executed.  At this point the plugin should load any config
        and perform any validations needed.
        The plugin should not create any resources but it is understood that
        there may be a cost of preparation.

        Provisioners are expected to load a lot of config to self-program.
        Because of this, they allow passing of a configerus label for .load()
        and a base for .get() in case there is an alterante config source
        desired.

        """
        raise NotImplementedError(
            'This provisioner has not yet implemented prepare')

    def apply(self):
        """Bring a cluster to the configured state."""
        raise NotImplementedError(
            'This provisioner has not yet implemented apply')

    def destroy(self):
        """Remove all resources created for the cluster."""
        raise NotImplementedError(
            'This provisioner has not yet implemented destroy')



"""

METTA Plugin definition and management.

The basis of plugin management is here.  The functionality allows any code to
register a plugin by decorating a factory function, or to use the factory to
ask for an instance of a plugin based on type an id.

Plugins have two important pieces of metadata for their type:

1. Plugin_type : an arbitrary string label which allows semantic grouping of
        plugins.  Type is used for loose retrieval, for example "to get all of
        the CLI plugin types", but is also used for introspection (checking
        out what we have.)
2. Plugin_Id : an arbitrary string used to label the specific plugin type.
        This means that when trying to get an instance of a specific plugin,
        this value can be used.

"""
import logging
import functools
from typing import Dict, Callable

logger = logging.getLogger('metta.plugin')

METTA_PLUGIN_CONFIG_KEY_PLUGIN = 'plugin'
""" configerus .get() key for a single plugin """
METTA_PLUGIN_CONFIG_KEY_PLUGINID = 'plugin_id'
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_INSTANCEID = 'instance_id'
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_PLUGINTYPE = 'plugin_type'
""" configerus .get() key for plugin type """
METTA_PLUGIN_CONFIG_KEY_ARGUMENTS = 'arguments'
""" configerus .get() key for plugin arguments """
METTA_PLUGIN_CONFIG_KEY_PRIORITY = 'priority'
""" configerus .get()  assign an instance a priority when it is created. """
METTA_PLUGIN_CONFIG_KEY_CONFIG = 'config'
""" configerus .get()  as additional config """
METTA_PLUGIN_CONFIG_KEY_VALIDATORS = 'validators'
""" configerus .get()  to decide what validators to apply to the plugin """


METTA_PLUGIN_VALIDATION_JSONSCHEMA = {
    METTA_PLUGIN_CONFIG_KEY_INSTANCEID: {'plugin_type': 'string'}
}
""" json schema validation definition for a plugin """


class Factory():
    """Python decorator class for metta Plugin factories.

    This class should be used to decorate any function which is meant to be a
    factory for metta plugins.

    If you are writing a plugin factory, decorate it with this class, provide
    the factory type and id values, and then the factory will be avaialble to
    other code.

    If you are trying to get an instance of a plugin, then create an instance
    of this class and use the create() method

    """

    registry: Dict[str, Dict[str, Callable]] = {}
    """ A dict of registered factory functions, grouped by plugin type """

    def __init__(self, plugin_type: str, plugin_id: str):
        """Create a new Factory instance.

        This is used in two scenarios:

        1. Execution of a decoration on a factory function.
        2. Initialization of an object that will be used to for new instances.

        Parameters:
        -----------
        plugin_type (str) : Plugin type label which the factory created
            pylint W0622 this is used only in this function and it makes for a
            much more natural decorator syntax

        plugin_id (str) : unique identifier for the plugin which this factory
            will create.  This will be used on registration and the matching
            value is used on construction.

        """
        logger.debug("Metta Plugin factory registering `%s:%s`", plugin_type, plugin_id)
        self.plugin_id = plugin_id
        self.plugin_type = plugin_type

        if self.plugin_type not in self.registry:
            self.registry[self.plugin_type] = {}

    def __call__(self, func):
        """Wrap the decorator factory wrapping function.

        Returns:
        --------
        Decorated construction function.

        """
        functools.update_wrapper(self, func)

        def wrapper(environment: object, instance_id: str, *args, **kwargs):
            logger.debug("plugin factory exec: %s:%s", self.plugin_type, self.plugin_id)
            plugin = func(
                environment=environment,
                instance_id=instance_id,
                *args,
                **kwargs)

            return plugin

        self.registry[self.plugin_type][self.plugin_id] = wrapper
        return wrapper

    def create(self, environment: object, instance_id: str, *args, **kwargs) -> object:
        """Get an instance of a plugin as created by the decorated.

        Parameters:
        -----------
        environment (Environment) : Environment object in which this plugin
            lives.

        instance_id (str) : instance id for the plugin.  A unique identifier
            which the plugin can use for naming and introspective
            identification.

        """
        try:
            factory = self.registry[self.plugin_type][self.plugin_id]
        except KeyError as err:
            raise NotImplementedError(
                f"METTA Plugin instance '{self.plugin_type}:{self.plugin_id}' "
                "has not been registered.") from err
        except Exception as err:
            raise Exception(
                f"Could not create Plugin instance '{self.plugin_type}:{self.plugin_id}' "
                "due to an unknown error.") from err

        return factory(environment=environment,
                       instance_id=instance_id, *args, **kwargs)

    def types(self):
        """Return a generator of types that have been registered."""
        return self.registry.keys()



"""

Metta Sonobuoy package.

Primarily we register the metta plugins

"""
from typing import Any

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.workload import METTA_PLUGIN_TYPE_WORKLOAD
from mirantis.testing.metta_cli.base import METTA_PLUGIN_TYPE_CLI

from .sonobuoy import (SonobuoyWorkloadPlugin, METTA_PLUGIN_ID_SONOBUOY_WORKLOAD,
                       SONOBUOY_WORKLOAD_CONFIG_LABEL, SONOBUOY_WORKLOAD_CONFIG_BASE)
from .cli import SonobuoyCliPlugin, METTA_PLUGIN_ID_SONOBUOY_CLI


@Factory(plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD)
def metta_plugin_factory_workload_sonobuoy(environment: Environment, instance_id: str = '',
                                           label: str = SONOBUOY_WORKLOAD_CONFIG_LABEL,
                                           base: Any = SONOBUOY_WORKLOAD_CONFIG_BASE):
    """Create an metta sonobuoy workload plugin."""
    return SonobuoyWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLI, plugin_id=METTA_PLUGIN_ID_SONOBUOY_CLI)
def metta_plugin_factory_cli_sonobuoy(environment: Environment, instance_id: str = ''):
    """Create an sonobuoy cli plugin."""
    return SonobuoyCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI : Environment commands.

Various commands that allow introspection of the available environments.

"""
import logging

from mirantis.testing.metta import environment_names, get_environment
from mirantis.testing.metta.environment import Environment

from .base import CliBase, cli_output

logger = logging.getLogger('metta.cli.environment')


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class EnvironmentCliPlugin(CliBase):
    """Fire command/group generator for environment commands."""

    def fire(self):
        """Return a dict of commands."""
        return {
            'environment': EnvironmentGroup(self.environment)
        }


class EnvironmentGroup():
    """ase Fire command group for environment commands."""

    def __init__(self, environment: Environment):
        """Create CLI command group."""
        self.environment = environment

    # needs to be a method for registration in fire
    # pylint: disable=no-self-use
    def list(self, raw: bool = False):
        """List all of the environment names."""
        names = environment_names()
        if raw:
            return names
        return cli_output(names)

    def _environment(self, environment: str = ''):
        """Select an environment."""
        if not environment:
            return self.environment
        return get_environment(environment)

    def name(self, environment: str = ''):
        """Return env name."""
        environment_object = self._environment(environment)
        return environment_object.name

    def info(self, environment: str = ''):
        """Return info about an environment."""
        environment_object = self._environment(environment)

        info = {
            'name': environment_object.name
        }

        if len(environment_object.states) > 0:
            info['states'] = {
                'available': environment_object.states,
                'active': environment_object.state
            }

        return cli_output(info)

    def bootstraps(self, environment: str = ''):
        """List bootstraps that have been applied to the environment."""
        environment_object = self._environment(environment)
        bootstrap_list = list(bootstrap for bootstrap in environment_object.bootstrapped)

        return cli_output(bootstrap_list)



"""

Terraform METTA provisioner plugin.

Provisioner plugin to provisioner cluster infrastructure using Terraform.

"""

import logging
import os
import subprocess
import shutil
from typing import Any

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import (Fixtures, METTA_FIXTURES_CONFIG_FIXTURES_LABEL)
from mirantis.testing.metta.provisioner import ProvisionerBase
from mirantis.testing.metta.output import METTA_PLUGIN_TYPE_OUTPUT
from mirantis.testing.metta_common import METTA_PLUGIN_ID_OUTPUT_DICT, METTA_PLUGIN_ID_OUTPUT_TEXT

from .terraform import TerraformClient

logger = logging.getLogger('metta_terraform:provisioner')

METTA_TERRAFORM_PROVISIONER_PLUGIN_ID = 'metta_terraform'
""" Terraform provisioner plugin id """

TERRAFORM_PROVISIONER_CONFIG_LABEL = 'terraform'
""" config label loading the terraform config """
TERRAFORM_PROVISIONER_CONFIG_ROOT_PATH_KEY = 'root.path'
""" config key for a base path that should be used for any relative paths """
TERRAFORM_PROVISIONER_CONFIG_PLAN_PATH_KEY = 'plan.path'
""" config key for the terraform plan path """
TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY = 'state.path'
""" config key for the terraform state path """
TERRAFORM_PROVISIONER_CONFIG_VARS_KEY = 'vars'
""" config key for the terraform vars Dict, which will be written to a file """
TERRAFORM_PROVISIONER_CONFIG_VARS_PATH_KEY = 'vars_path'
""" config key for the terraform vars file path, where the plugin will write to """
TERRAFORM_PROVISIONER_DEFAULT_VARS_FILE = 'metta_terraform.tfvars.json'
""" Default vars file if none was specified """
TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH = 'metta-state'
""" Default vars file if none was specified """

TERRAFORM_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'type': {'type': 'string'},
        'plugin_id': {'type': 'string'},

        'root': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            }
        },
        'plan': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            },
            'required': ['path']
        },
        'state': {
            'type': 'object',
            'properties': {
                'path': {'type': 'string'}
            }
        },
        'vars_path': {
            'type': 'string'
        },
        'vars': {
            'type': 'object'
        }
    }
}
""" Validation jsonschema for terraform config contents """
TERRAFORM_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: TERRAFORM_VALIDATE_JSONSCHEMA}
""" configerus validation target to match the above config """


# pylint: disable=too-many-instance-attributes
class TerraformProvisionerPlugin(ProvisionerBase):
    """Terraform provisioner plugin.

    Provisioner plugin that allows control of and interaction with a terraform
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a terraform binary, so you have to
        install terraform in the environment

    ## Usage

    ### Plan

    The plan must exists somewhere on disk, and be accessible.

    You must specify the path and related configuration in config, which are
    read in the .prepare() execution.

    ### Vars/State

    This plugin reads TF vars from config and writes them to a vars file.  We
    could run without relying on vars file, but having a vars file allows cli
    interaction with the cluster if this plugin messes up.

    You can override where Terraform vars/state files are written to allow
    sharing of a plan across test suites.

    Parameters:
    -----------
    environment (Environment) : All metta plugins receive the environment
        object in which they were created.
    instance_id (str) : all metta plugins receive their own string identity.

    label (str) : Configerus load label for finding plugin config.
    base (str) : Configerus get base key in which the plugin should look for
        all config.

    """

    def __init__(self, environment: Environment, instance_id: str,
                 label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """Run the super constructor but also set class properties.

        Interpret provided config and configure the object with all of the
        needed pieces for executing terraform commands

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        logger.info("Preparing Terraform setting")

        self.terraform_config_label = label
        """ configerus load label that should contain all of the config """
        self.terraform_config_base = base
        """ configerus get key that should contain all tf config """

        terraform_config = self.environment.config.load(self.terraform_config_label)
        """ get a configerus LoadedConfig for the terraform label """

        # Run confgerus validation on the config using our above defined
        # jsonschema
        try:
            terraform_config.get(self.terraform_config_base, validator=TERRAFORM_VALIDATE_TARGET)
        except ValidationError as err:
            raise ValueError("Terraform config failed validation") from err

        self.fixtures = self.environment.add_fixtures_from_config(
            label=self.terraform_config_label,
            base=[self.terraform_config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL])
        """ All fixtures added to this provisioner plugin. """

        self.root_path = terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_ROOT_PATH_KEY], default='')
        """ all relative paths will have this joined as their base """

        try:
            self.working_dir = terraform_config.get(
                [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_PLAN_PATH_KEY])
            """ subprocess commands for terraform will be run in this path """
        except Exception as err:
            raise ValueError("Plugin config did not give us a working/plan path:"
                             f" {terraform_config.data}") from err
        if not os.path.isabs(self.working_dir):
            if self.root_path:
                self.working_dir = os.path.join(self.root_path, self.working_dir)
            self.working_dir = os.path.abspath(self.working_dir)

        state_path = terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY],
            default=os.path.join(self.working_dir, TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH))
        """ terraform state path """
        if not os.path.isabs(state_path):
            if self.root_path:
                state_path = os.path.join(self.root_path, state_path)
            state_path = os.path.abspath(state_path)

        self.vars = terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_VARS_KEY], default={})
        """ List of vars to pass to terraform.  Will be written to a file """

        vars_path = terraform_config.get(
            [self.terraform_config_base, TERRAFORM_PROVISIONER_CONFIG_VARS_PATH_KEY],
            default=os.path.join(self.working_dir, TERRAFORM_PROVISIONER_DEFAULT_VARS_FILE))
        """ vars file which will be written before running terraform """
        if not os.path.isabs(vars_path):
            if self.root_path:
                vars_path = os.path.join(self.root_path, vars_path)
            vars_path = os.path.abspath(vars_path)

        logger.info("Creating Terraform client")

        self.tf_client = TerraformClient(
            working_dir=os.path.realpath(self.working_dir),
            state_path=os.path.realpath(state_path),
            vars_path=os.path.realpath(vars_path),
            variables=self.vars)
        """ TerraformClient instance """

        # if the cluster is already provisioned then we can get outputs from it
        try:
            self._get_outputs_from_tf()
        except subprocess.CalledProcessError:
            pass

    def info(self):
        """Get info about a provisioner plugin.

        Returns:
        --------
        Dict of keyed introspective information about the plugin.

        """
        plugin = self
        client = plugin.tf_client

        info = {}
        info['plugin'] = {
            'terraform_config_label': plugin.terraform_config_label,
            'terraform_config_base': plugin.terraform_config_base
        }
        info['client'] = {
            'vars': client.vars,
            'working_dir': client.working_dir,
            'state_path': client.state_path,
            'vars_path': client.vars_path,
            'terraform_bin': client.terraform_bin
        }

        fixtures = {}
        for fixture in self.fixtures:
            fixture_info = {
                'fixture': {
                    'plugin_type': fixture.plugin_type,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }
            if hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    fixture_info.update(plugin_info)
            fixtures[fixture.instance_id] = plugin_info
        info['fixtures'] = fixtures

        # build a list of helpful terraform command strings to allow users to directly run
        # terraform as we would, with the metta vars/state/working dir
        args = {'bin': client.terraform_bin, 'working_dir': client.working_dir,
                'vars_path': client.vars_path, 'state_path': client.state_path}
        info['helper'] = {
            'commands': {
                'init': f"{args['bin']} -chdir={args['working_dir']} init",
                'plan': f"{args['bin']} -chdir={args['working_dir']} plan "
                        f"-var-file={args['vars_path']} -state={args['state_path']}",
                'apply': f"{args['bin']} -chdir={args['working_dir']} apply "
                         f"-var-file={args['vars_path']} -state={args['state_path']}",
                'destroy': f"{args['bin']} -chdir={args['working_dir']} destroy "
                           f"-var-file={args['vars_path']} -state={args['state_path']}",
                'output': f"{args['bin']} -chdir={args['working_dir']} output "
                          f"-state={args['state_path']}"
            }
        }

        return info

    def prepare(self):
        """Run terraform init."""
        logger.info("Running Terraform INIT")
        self.tf_client.init()

    def check(self):
        """Check that the terraform plan is valid."""
        logger.info("Running Terraform PLAN")
        self.tf_client.plan()

    def apply(self):
        """Create all terraform resources described in the plan."""
        logger.info("Running Terraform APPLY")
        self.tf_client.apply()
        self._get_outputs_from_tf()

    def destroy(self):
        """Remove all terraform resources in state."""
        logger.info("Running Terraform DESTROY")
        self.tf_client.destroy()
        # accessing parent property for clearing out existing output fixtures
        # pylint: disable=attribute-defined-outside-init
        self.fixtures = Fixtures()

    def clean(self):
        """Remove terraform run resources from the plan."""
        logger.info("Running Terraform CLEAN")
        dot_terraform = os.path.join(self.working_dir, '.terraform')
        if os.path.isdir(dot_terraform):
            shutil.rmtree(dot_terraform)

    # ----- Cluster Interaction -----

    def _get_outputs_from_tf(self):
        """Retrieve an output from terraform.

        For other METTA plugins we can just load configuration, and creating
        output plugin instances from various value in config.

        We do that here, but  we also want to check of any outputs exported by
        the terraform root module, which we get using the tf client.

        If we find a root module output without a matching config output
        defintition then we make some assumptions about plugin type and add it
        to the list. We make some simple investigation into output plugin types
        and pick either the contrib.common.dict or contrib.common.text plugins.

        If we find a root module output that matches an output that was
        declared in config then we use that.  This allows config to define a
        plugin_id which will then be populated automatically.  If you know what
        type of data you are expecting from a particular tf output then you can
        prepare and config for it to do things like setting default values.

        Priorities can be used in the config.

        """
        # now we ask TF what output it nows about and merge together those as
        # new output plugins.
        # tf.outputs() produces a list of (sensitive:bool, type: [str,  object,
        # value:Any])
        for output_key, output_struct in self.tf_client.output().items():

            # Here is the kind of info we can get out of terraform
            # output_sensitive = bool(output_struct['sensitive'])
            # """ Whether or not the output contains sensitive data """
            output_type = output_struct['type'][0]
            """ output primitive type (usually string|object|number) """
            # output_spec = output_struct['type'][1]
            # """ A structured spec for the type """
            output_value = output_struct['value']
            """ output value """

            # see if we already have an output plugin for this name
            fixture = self.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_OUTPUT,
                instance_id=output_key,
                exception_if_missing=False)
            if not fixture:
                # we only know how to create 2 kinds of outputs
                if output_type == 'object':
                    fixture = self.environment.add_fixture(
                        plugin_type=METTA_PLUGIN_TYPE_OUTPUT,
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_DICT,
                        instance_id=output_key,
                        priority=self.environment.plugin_priority(delta=5),
                        arguments={'data': output_value})
                else:
                    fixture = self.environment.add_fixture(
                        plugin_type=METTA_PLUGIN_TYPE_OUTPUT,
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_TEXT,
                        instance_id=output_key,
                        priority=self.environment.plugin_priority(delta=5),
                        arguments={'text': str(output_value)})

                self.fixtures.add(fixture)
            else:
                if hasattr(fixture.plugin, 'set_data'):
                    fixture.plugin.set_data(output_value)
                elif hasattr(fixture.plugin, 'set_text'):
                    fixture.plugin.set_text(str(output_value))



"""

PLUGIN: Output.

Workload plugins use clients to apply workloads to a testing system.  The
system is irrelevant tothe workload plugin, as it only cares whether or not it
can retrieve the expected clients.

Workload clients are expected to allow creation of Workload instances from
clients, so that multiple applications of the workload can be managed at one
time.

"""

import logging

from .fixtures import Fixtures

logger = logging.getLogger('metta.workload')

METTA_PLUGIN_TYPE_WORKLOAD = "workload"
""" Fast access to the workload type """

METTA_WORKLOAD_CONFIG_WORKLOADS_LABEL = 'workloads'
""" A centralized configerus load label for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL = 'workload'
""" A centralized configerus load label for a workload """
METTA_WORKLOAD_CONFIG_WORKLOADS_KEY = 'workloads'
""" A centralized configerus key for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_KEY = 'workload'
""" A centralized configerus key for one workload """


# pylint: disable=too-few-public-methods
class WorkloadBase:
    """Base class for workload plugins."""

    def create_instance(self, fixtures: Fixtures) -> 'WorkloadInstanceBase':
        """Create a workload instance from a set of fixtures."""


class WorkloadInstanceBase:
    """An instance of a running workload."""

    def apply(self):
        """Run the workload.

        @NOTE Needs a kubernetes client fixture to run.
              Use .set_fixtures() first

        """
        raise NotImplementedError(
            "This workload plugin has not implemented apply()")

    def destroy(self):
        """Destroy any created resources."""
        raise NotImplementedError(
            "This workload plugin has not implemented destroy()")



"""

Run a Sonobuoy run on a k82 client.

Use this to run the sonobuoy implementation

"""
from typing import Dict, Any, List
import logging
import subprocess
import os
import json
from enum import Enum, unique

import yaml

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase
from mirantis.testing.metta_kubernetes import (METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                                               KubernetesApiClientPlugin)

logger = logging.getLogger('workload.sonobuoy')

METTA_PLUGIN_ID_SONOBUOY_WORKLOAD = 'metta_sonobuoy_workload'
""" workload plugin_id for the sonobuoy plugin """

SONOBUOY_WORKLOAD_CONFIG_LABEL = 'sonobuoy'
""" Configerus label for retrieving sonobuoy config """
SONOBUOY_WORKLOAD_CONFIG_BASE = LOADED_KEY_ROOT
""" Configerus get base for retrieving the default workload config """

SONOBUOY_CONFIG_KEY_MODE = 'mode'
""" config key for mode """
SONOBUOY_CONFIG_KEY_KUBERNETESVERSION = 'kubernetes.version'
""" config key for kubernetes version """
SONOBUOY_CONFIG_KEY_PLUGINS = 'plugin.plugins'
""" config key for what plugins to run """
SONOBUOY_CONFIG_KEY_PLUGINENVS = 'plugin.envs'
""" config key for plugin env flags """

SONODBUOY_DEFAULT_WAIT_PERIOD_SECS = 1440
""" Default time for sonobuoy to wait when running """

SONOBUOY_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'type': {'type': 'string'},
        'plugin_id': {'type': 'string'},

        'mode': {
            'type': 'string'
        },
        'kubernetes': {
            'type': 'object',
            'properties': {
                'version': {'type': 'string'}
            },
        },
        'kubernetes_version': {
            'type': 'string'
        },
        'plugin': {
            'type': 'object',
            'properties': {
                'plugins': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                },
                'plugin_envs': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                }
            }
        },
    },
    'required': ['mode', 'kubernetes']
}
""" Validation jsonschema for terraform config contents """
SONOBUOY_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: SONOBUOY_VALIDATE_JSONSCHEMA}
""" configerus validation target to match the jsonschema config """


class SonobuoyWorkloadPlugin(WorkloadBase):
    """Workload class for the Sonobuoy"""

    def __init__(self, environment: Environment, instance_id: str,
                 label: str = SONOBUOY_WORKLOAD_CONFIG_LABEL,
                 base: Any = SONOBUOY_WORKLOAD_CONFIG_BASE):
        """Initialize workload plugin.

        Parameters:
        -----------
        label (str) : Configerus label for loading config
        base (Any) : configerus base key which should contain all of the config

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        logger.info("Preparing sonobuoy settings")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes api client plugin.

        """
        loaded = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the sonobuoy label """

        # Validate the config overall using jsonschema
        try:
            loaded.get(self.config_base, validator=SONOBUOY_VALIDATE_TARGET)
        except ValidationError as err:
            raise ValueError("Invalid sonobuoy config received") from err

        kubeclient = fixtures.get_plugin(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)

        mode = loaded.get([self.config_base,
                           SONOBUOY_CONFIG_KEY_MODE])
        kubernetes_version = loaded.get(
            [self.config_base, SONOBUOY_CONFIG_KEY_KUBERNETESVERSION], default='')
        plugins = loaded.get(
            [self.config_base, SONOBUOY_CONFIG_KEY_PLUGINS], default=[])
        plugin_envs = loaded.get(
            [self.config_base, SONOBUOY_CONFIG_KEY_PLUGINENVS], default=[])

        return SonobuoyConformanceWorkloadInstance(
            kubeclient=kubeclient, mode=mode, kubernetes_version=kubernetes_version,
            plugins=plugins, plugin_envs=plugin_envs)

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        loaded = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the sonobuoy label """

        sonobuoy_config = loaded.get(self.config_base)

        kubeclient = self.environment.fixtures.get_plugin(
            plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT)

        info = {
            'workload': {
                'cncf': sonobuoy_config,
                'kube_client': kubeclient.info(),
                'required_fixtures': {
                    'kubernetes': {
                        'plugin_type': METTA_PLUGIN_TYPE_CLIENT,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }

        return info


SONOBUOY_DEFAULT_BIN = 'sonobuoy'
""" Default Bin Name for running sonobuoy """
SONOBUOY_DEFAULT_RESULTS_PATH = './results'
""" Default path for where to download sonobuoy results """


class SonobuoyConformanceWorkloadInstance(WorkloadInstanceBase):
    """A conformance workload instance for a docker run."""

    # pylint: disable=too-many-arguments
    def __init__(self, kubeclient: KubernetesApiClientPlugin, mode: str,
                 kubernetes_version: str = '', plugins: List[str] = None,
                 plugin_envs: List[str] = None, binary: str = SONOBUOY_DEFAULT_BIN,
                 results_path: str = SONOBUOY_DEFAULT_RESULTS_PATH):
        """Imitialize the workload instance."""
        self.kubeclient = kubeclient
        """ metta kube client, which gives us a kubeconfig """
        self.mode = mode
        """ sonobuoy mode, passed to the cli """
        self.kubernetes_version = kubernetes_version
        """ Kubernetes version to test compare against """
        self.plugins = plugins if plugins is not None else []
        """ which sonobuoy plugins to run """
        self.plugin_envs = plugin_envs if plugin_envs is not None else []
        """ Plugin specific ENVs to pass to sonobuoy """

        self.bin = binary
        """ path to the sonobuoy binary """

        self.results_path = results_path
        """ path to where to download sonobuoy results """

    # These are work for this scenario
    # pylint: disable=arguments-differ
    def apply(self, wait: bool = True):
        """Run sonobuoy."""
        cmd = ['run']
        # we don't need to add --kubeconfig here as self._run() does that

        if self.mode:
            cmd += [f'--mode={self.mode}']

        if self.kubernetes_version:
            cmd += [f'--kube-conformance-image-version={self.kubernetes_version}']

        if self.plugins:
            cmd += [f'--plugin={plugin_id}' for plugin_id in self.plugins]

        if self.plugin_envs:
            cmd += [f'--plugin-env={plugin_env}' for plugin_env in self.plugin_envs]

        if wait:
            cmd += [f'--wait={SONODBUOY_DEFAULT_WAIT_PERIOD_SECS}']

        logger.info("Starting Sonobuoy run : %s", cmd)
        try:
            self._create_k8s_crb()
            self._run(cmd)
        except Exception as err:
            raise RuntimeError("Sonobuoy RUN failed") from err

    def status(self):
        """Retrieve Sonobuoy status return."""
        cmd = ['status', '--json']
        status = self._run(cmd, return_output=True)
        if status:
            return SonobuoyStatus(status)

        return None

    def logs(self, follow: bool = True):
        """Retrieve sonobuoy logs."""
        cmd = ['logs']

        if follow:
            cmd += ['--follow']

        self._run(cmd)

    def retrieve(self):
        """Retrieve sonobuoy results."""
        logger.debug("retrieving sonobuoy results")
        try:
            cmd = ['retrieve', self.results_path]
            file = self._run(cmd=cmd, return_output=True).rstrip("\n")
            if not os.path.isfile(file):
                raise RuntimeError('Sonobuoy did not retrieve a results tarball.')
            return SonobuoyResults(tarball=file, folder=self.results_path)
        except Exception as err:
            raise RuntimeError("Could not retrieve sonobuoy results") from err

    # These are work for this scenario
    # pylint: disable=arguments-differ
    def destroy(self, wait: bool = False):
        """Delete sonobuoy resources."""
        cmd = ['delete']

        if wait:
            cmd += ['--wait']

        self._run(cmd)
        self._delete_k8s_crb()

    def _run(self, cmd: List[str], ignore_errors: bool = True, return_output: bool = False):
        """Run a sonobuoy command."""
        kubeconfig = self.kubeclient.config_file
        cmd = [self.bin, f'--kubeconfig={kubeconfig}'] + cmd

        # this else makes it much more readable
        # pylint: disable=no-else-return
        if return_output:
            logger.debug(
                "running sonobuoy command with output capture: %s",
                " ".join(cmd))
            res = subprocess.run(cmd, shell=False, check=True, stdout=subprocess.PIPE)

            # sonobuoy's uses of subprocess error is overly inclusive for us
            if not ignore_errors:
                res.check_returncode()

            return res.stdout.decode('utf-8')

        else:
            logger.debug("running sonobuoy command: %s", " ".join(cmd))
            res = subprocess.run(cmd, check=True, text=True)

            if not ignore_errors:
                res.check_returncode()

            return res

    def _create_k8s_crb(self):
        """Create the cluster role binding that sonobuoy needs."""
        # Sonobuoy requires an admin cluster role binding
        # @TODO Do this with the kubectl client properly
        crb_cmds = [
            'kubectl',
            'create',
            'clusterrolebinding',
            'sonobuoy-serviceaccount-cluster-admin',
            '--clusterrole=cluster-admin',
            '--serviceaccount=sonobuoy:sonobuoy-serviceaccount']
        env = os.environ.copy()
        env['KUBECONFIG'] = self.kubeclient.config_file
        logger.debug(
            "creating sonobuoy cluster_role_binding: %s",
            " ".join(crb_cmds))
        try:
            res = subprocess.run(crb_cmds, env=env, check=True, text=True)
            res.check_returncode()
            return res
        except subprocess.CalledProcessError:
            # this typically means that the CRBs already exist
            return None

    def _delete_k8s_crb(self):
        """Remove the cluster role binding that we created."""
        # @TODO Do this with the kubectl client properly
        crb_cmds = [
            'kubectl',
            'delete',
            'clusterrolebinding',
            'sonobuoy-serviceaccount-cluster-admin']
        env = os.environ.copy()
        env['KUBECONFIG'] = self.kubeclient.config_file
        logger.debug(
            "creating sonobuoy cluster_role_binding: %s",
            " ".join(crb_cmds))
        res = subprocess.run(
            crb_cmds, env=env, check=True, text=True)
        res.check_returncode()
        return res


class SonobuoyStatus:
    """A status ooutput from the sonobuoy CLI."""

    def __init__(self, status_json: str):
        """Build from sonobuoy status results."""
        status = json.loads(status_json)
        self.status = Status(status['status'])
        self.tar_info = status['tar-info']

        self.plugins = {}
        for plugin in status['plugins']:
            self.plugins[plugin['plugin']] = plugin

    def plugin_list(self):
        """Retrieve the list of plugins."""
        return list(self.plugins.keys())

    def plugin(self, plugin: str):
        """Retrieve the results for one plugin."""
        return self.plugins[plugin]

    def plugin_status(self, plugin: str):
        """Get the status code for a plugin."""
        status_string = self.plugin(plugin)['status']
        return Status(status_string)


class SonobuoyResults:
    """Results retrieved analyzer."""

    def __init__(self, tarball: str, folder: str):
        """Interpret tarball contents."""
        logger.debug("un-tarring retrieved results: %s", tarball)
        res = subprocess.run(['tar', '-xzf', tarball, '-C', folder], check=True, text=True)
        res.check_returncode()

        self.results_path = folder

        with open(os.path.join(folder, 'meta', 'config.json')) as config_json:
            self.meta_config = json.load(config_json)
        with open(os.path.join(folder, 'meta', 'info.json')) as info_json:
            self.meta_info = json.load(info_json)
        with open(os.path.join(folder, 'meta', 'query-time.json')) as qt_json:
            self.meta_querytime = json.load(qt_json)

        self.plugins = []
        for plugin_id in self.meta_info['plugins']:
            self.plugins.append(plugin_id)

    def plugin_list(self):
        """Return a string list of plugin ids."""
        return self.plugins

    def plugin(self, plugin_id):
        """Return the results for a single plugin."""
        return SonobuoyResultsPlugin(os.path.join(
            self.results_path, 'plugins', plugin_id))


class SonobuoyResultsPlugin:
    """The full results for a plugin."""

    def __init__(self, path: str):
        """Load results for a plugin results call."""
        with open(os.path.join(path, 'sonobuoy_results.yaml')) as results_yaml:
            self.summary = yaml.safe_load(results_yaml)

    def name(self) -> str:
        """Return string name of plugin."""
        return self.summary['name']

    def status(self) -> "Status":
        """Return the status object for the plugin results."""
        return Status(self.summary['status'])

    def __len__(self):
        """Count how many items are in the plugin_results."""
        return len(self.summary['items'])

    def __getitem__(self, instance_id: Any) -> object:
        """Get item details from the plugin results."""
        return SonobuoyResultsPluginItem(
            item_dict=self.summary['items'][instance_id])


class SonobuoyResultsPluginItem:
    """An individual item from a sonobuoy results plugin."""

    def __init__(self, item_dict: Dict[str, Any]):
        """Single plugin result item."""
        self.name = item_dict['name']
        self.status = Status(item_dict['status'])
        self.meta = item_dict['meta']
        self.details = item_dict['details'] if 'details' in item_dict else {}

    def meta_file_path(self):
        """Get the path to the error item file."""
        return self.meta['file']

    def meta_file(self):
        """Get the contents of the file."""
        with open(self.meta_file_path()) as meta_file:
            return yaml.safe_load(meta_file)


@unique
class Status(Enum):
    """Enumerator to plugin states."""

    PENDING = 'pending'
    """ still pending """
    RUNNING = 'running'
    """ testing is running """
    FAILED = 'failed'
    """ testing has failed """
    COMPLETE = 'complete'
    """ testing has completed without failure """
    PASSED = 'passed'
    """ testing has passed """
    POSTPROCESS = 'post-processing'
    """ testing has finished and is being processed """



"""

Configerus extension to connect metta output plugins to configerus formatting.

This allows embedding of metta output content into configerus configuration.

"""
import re
import logging

from configerus.config import Config

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.output import METTA_PLUGIN_TYPE_OUTPUT
from mirantis.testing.metta_common.dict_output import DictOutputPlugin
from mirantis.testing.metta_common.text_output import TextOutputPlugin

logger = logging.getLogger('configerus.contrib.files:output')


OUTPUT_FORMAT_MATCH_PATTERN = r'(?P<output>(\w+)+)(\/(?P<base>[\-\.\w]+))?'
""" A regex pattern to identify outputs that should be embedded """


PLUGIN_ID_FORMAT_OUTPUT = 'output'
""" Format plugin_id for the configerus output format plugin """


class ConfigFormatOutputPlugin:
    """ Configerus formatter plugin that uses Metta outputs as a source """

    def __init__(self, config: Config, instance_id: str):
        """Create configerus format plugin."""
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(OUTPUT_FORMAT_MATCH_PATTERN)
        """ Regex patter for identifying an output replacement """

        self.environment: Environment = None
        """ environment which contains the outputs. Must be added """

    def set_environemnt(self, environment: Environment):
        """Set the output environment.

        @NOTE this is obligatory

        """
        self.environment = environment

    # this method is a part of an itnerface. default label is not used by us but it will be passed
    # pylint: disable=unused-argument
    def format(self, key, default_label: str):
        """Format a string by substituting config values.

        Parameters
        ----------
        key: a string that should gies instructions to the formatter on how to
            create a format replacementsrmed

        default_source : if format/replace patterns don't have a source defined
            then this is used as a source.

        """
        # if the entire key is the match, then replace whatever type we get
        # out of the config .get() call
        match = self.pattern.fullmatch(key)
        if not match:
            raise KeyError(f"Could not interpret Format action key '{key}'")

        output = match.group('output')

        try:
            output_plugin = self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_OUTPUT, instance_id=output).plugin

            if isinstance(output_plugin, DictOutputPlugin):
                base = match.group('base')
                if base is not None:
                    return output_plugin.get_output(base)
                return output_plugin.get_output()
            if isinstance(output_plugin, TextOutputPlugin):
                return output_plugin.get_output()
            if hasattr(output_plugin, 'get_output'):
                return output_plugin.get_output()

            return ''

        except KeyError as err:
            raise KeyError(f"Config replace for output failed as output '{output}' was not found, "
                           "and no default value was suggested.") from err



"""

METTA Dummy.

Dummy plugin functionality.  Various plugins that can be used as placeholders
and for testing.

"""

import logging
from typing import Any, Dict

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_TYPE_PROVISIONER
from mirantis.testing.metta.client import METTA_PLUGIN_TYPE_CLIENT
from mirantis.testing.metta.workload import METTA_PLUGIN_TYPE_WORKLOAD

from .provisioner import DummyProvisionerPlugin
from .client import DummyClientPlugin
from .workload import DummyWorkloadPlugin

logger = logging.getLogger('metta.dummy')

METTA_PLUGIN_ID_DUMMY = 'dummy'
""" All of the dummy plugins use 'dummy' as their plugin_id """


@Factory(plugin_type=METTA_PLUGIN_TYPE_PROVISIONER, plugin_id=METTA_PLUGIN_ID_DUMMY)
def metta_plugin_factory_provisioner_dummy(environment: Environment, instance_id: str = '',
                                           fixtures: Dict[str, Dict[str, Any]] = None):
    """Create an metta provisionersss dict plugin."""
    return DummyProvisionerPlugin(environment, instance_id, fixtures)


@Factory(plugin_type=METTA_PLUGIN_TYPE_CLIENT, plugin_id=METTA_PLUGIN_ID_DUMMY)
def metta_plugin_factory_client_dummy(environment: Environment, instance_id: str = '',
                                      fixtures: Dict[str, Dict[str, Any]] = None):
    """Create an metta client dict plugin."""
    return DummyClientPlugin(environment, instance_id, fixtures)


@Factory(plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id=METTA_PLUGIN_ID_DUMMY)
def metta_plugin_factory_workload_dummy(environment: Environment, instance_id: str = '',
                                        fixtures: Dict[str, Dict[str, Any]] = None):
    """Create an metta workload dict plugin."""
    return DummyWorkloadPlugin(environment, instance_id, fixtures)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """
